/****************************************************************************

FILE	gener.c

PURPOSE
    This file contains auxilary functions, which are used mainly by
    grammar file.

*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "shade.h"
#include "opcodes.h"
#include "shmem.h"
#include "shdrglob.h"
#include "gener.h"
#include "y.tab.h"

int warning_flag=0;

void yyerror(char *s)
{
    fprintf(stderr, "<E>file %s: line %d: %s\n", cur_input_file,lineno, s);
    exit(1);
}

void undeclared(char *s)
{
    fprintf(stderr, "<E>file %s: line %d: Undeclared indentifier: %s\n", 
            cur_input_file,lineno, s);
    exit(1);
}


void warning(const char*text)
{
    if (warning_flag){
        fprintf(stderr, "<W> file %s: line %d: %s\n", cur_input_file, lineno, text);
    }
}

char * strsave(char *s)
{
    char *p;
    p = (char *)sh_malloc(strlen(s)+1);
    strcpy(p, s);
    return p;
}

/* automatic type conversion between expression trees for binary op */
struct tree * arithnode(int op, struct tree *l, struct tree *r)
{
    if (l->ty == tCOLOR) {
        if (r->ty == tCOLOR)
            return node2(op, tCOLOR, l, r);
        if (r->ty == tFLOAT)
            return node2(op, tCOLOR, l,
                node2(tCVTF2TRIPLE, tCOLOR, r, NULL));
    }
    else if ( isgeomtriple(l->ty) ) {
        if ( isgeomtriple(r->ty) )
            return node2(op, l->ty, l, r);
        if (r->ty == tFLOAT)
            return node2(op, l->ty, l,
                node2(tCVTF2TRIPLE, l->ty, r, NULL));
    }
    else if (l->ty == tFLOAT) {
        if (r->ty == tCOLOR)
            return node2(op, tCOLOR,
                node2(tCVTF2TRIPLE, tCOLOR, l, NULL), r);
        else if ( isgeomtriple( r->ty) )
            return node2(op, r->ty,
                node2(tCVTF2TRIPLE, r->ty, l, NULL), r);
        else if (r->ty == tFLOAT)
            return node2(op, tFLOAT, l, r);
    }
    return NULL;
}

/* automatic type conversion */
int assigntype(int lty, int rty)
{
    if ( istriple(lty) ) {
        if (rty == tFLOAT)
            return lty; /* floats convert to these */
        if (lty ==tCOLOR && rty == tCOLOR ) /* color to color */
            return lty;
        if (isgeomtriple (lty) && isgeomtriple(rty) )
            return lty;
    } 
    else if (lty == tFLOAT && rty == tFLOAT)
        return tFLOAT;
    else if (lty == tSTRING && rty == tSTRING)
        return tSTRING;
    return -1;
}

/* Can == or != be performed on these types? */
struct tree * relequiv(int op, int op3, struct tree *l, struct tree *r)
{
    int tmp_op = -1;
    if ( istriple( l->ty) && r->ty == tFLOAT) {
        tmp_op = op3;
        r = node2(tCVTF2TRIPLE, l->ty, r, NULL);
    }
    else if (l->ty == tFLOAT && istriple(r->ty )) {
        tmp_op = op3;
        l = node2(tCVTF2TRIPLE, r->ty, l, NULL);
    } 
    else if (l->ty == tFLOAT && r->ty== tFLOAT ) {
        tmp_op = op;
    } 
    else if (l->ty == tCOLOR && r->ty == tCOLOR){
        tmp_op = op3;
    }
    else if ( isgeomtriple(l->ty) && isgeomtriple(r->ty)){
        tmp_op = op3; 
    }
    else if (l->ty == tSTRING && r->ty == tSTRING){
        tmp_op = op;
    }
    return node2(tmp_op, l->ty, l, r);
    
}

int istriple(int type)
{
  return type==tCOLOR || type==tPOINT || type==tNORMAL || type==tVECTOR;
}

int isgeomtriple(int type)
{
  return type==tPOINT || type==tNORMAL || type==tVECTOR;
}


struct tree * relop(int op, struct tree *l, struct tree *r)
{
    if (l->ty != tFLOAT || r->ty != tFLOAT)
        return NULL;
    return node2(op, l->ty, l, r);
}


struct tree * node2(int op, int ty, struct tree *l, struct tree *r)
{
    struct tree *p;
    static int ind = 0;
    int n;
    p = (struct tree *)sh_malloc(sizeof(struct tree));
    p->op = op;
    p->ty = ty;
    p->off = -1;
    p->next_tree = NULL;
    p->ind = ind++;
    p->kids[0] = l;
    p->kids[1] = r;
    for(n = 2; n < MAXKIDS; n++)
        p->kids[n] = NULL;
    return p;
}

struct tree * node3(int op, int ty, struct tree *p0, struct tree *p1, struct tree *p2)
{
    struct tree *p;
    p = node2(op, ty, p0, p1);
    p->kids[2] = p2;
    return p;
}

struct tree * node4(int op, int ty, struct tree *p0, struct tree *p1, struct tree *p2, struct tree *p3)
{
    struct tree *p;
    p = node3(op, ty, p0, p1, p2);
    p->kids[3] = p3;
    return p;
}

struct tree * constnode(int ty, int offset)
{
    struct tree *p;
    p = node2(ty, ty, NULL, NULL);
    p->off = offset;
    return p;
}


struct ilist * merge(struct ilist *p1, struct ilist *p2)
{
    struct ilist *p = p1;
    if (p == NULL)
        return p2;
    while (p->next_ilist != NULL)
        p = p->next_ilist;
    p->next_ilist = p2;
    return p1;
}

struct ilist * mkilist(int n)
{
    struct ilist *p;
    p = (struct ilist *)sh_malloc(sizeof(struct ilist));
    p->i = n;
    p->next_ilist = NULL;
    return p;
}

void backpatch(struct ilist *t, int n)
{
    while (t != NULL) {
        storeshort(t->i, n);
        t = t->next_ilist;
    }
}

/* returns expression list length in bytes  */
int tlen(struct tree *t)
{
    int l = 0;
    while (t != NULL) {
        if (t->kids[1] != NULL) {
            if (t->kids[1]->ty == tFLOAT)
                l++;
            else if (istriple(t->kids[1]->ty))
                l += 3;
        }
        t = t->kids[0];
    }
    return l;
}

void use_sym(struct _sym *s)
{
    s->flags |= S_REFERENCED;
}

/* 
    checks function formal and actual parameter types
    generates expressions for function calling, if needed 
*/
void check_params(struct func_tree *function, struct tree *param_list)
{
    struct tree *act_param = param_list;
    struct func_param *formal_param = function->first_param;
    while (act_param != NULL && formal_param != NULL){
        struct tree *exprnode = act_param->kids[1];
        if (assigntype(formal_param ->param->type, act_param->ty) != 
                formal_param ->param->type){
            yyerror("parameter types don't match");
        }
        if (exprnode->off>=0 &&  !is_constant(exprnode->off) ){
            /* variable is sent as actual parameter */
            if (exprnode->op == tINDEX ||
                exprnode->op == tINDEX3){
                /* array element expression */
                formal_param->array_index = exprnode;
            }
            else{
                formal_param->array_index = NULL;
            }
            /* replace references in function with actual parameter ones */
            formal_param->param->repl_offset = exprnode->off;
            if (formal_param->param->flags&S_OUTPUT){
                /* 
                    check, whether we pass non-ouput function parameter to
                    inner function as output parameter
                */
                struct _sym * param = getsym(exprnode->off);
                int flags = param->flags;
                if ((flags & F_PARAM) && !(flags&S_OUTPUT)){
                    yyerror("passing non-modifiable parameter as output");
                }
            }
        }
        else if(formal_param->param->flags & S_REFERENCED){
            /* emit expression tree for parameter */
            struct tree *tmp=node2(act_param->op, act_param->ty, 
                NULL, act_param->kids[1]);
            struct tree *tr;
            formal_param->param->repl_offset = -1;                
            if (istriple(formal_param->param->type) && act_param->ty == tFLOAT){
                tmp = node2(tCVTF2TRIPLE, formal_param->param->type, tmp, NULL);
            }
            tr=node2('P', formal_param->param->type, 
                constnode(formal_param->param->type, formal_param->param->offset), tmp);
            emit(tr);
        }
        act_param = act_param->kids[0];
        formal_param = formal_param -> next_param;
    }
    if (act_param != NULL || formal_param != NULL){
        yyerror("formal and actual parameter count does not match");
    }
}


int  gen_arr_initializer(int type, int dim, int init_offs, struct tree * initializers)
{

    struct tree *cur_node = initializers;
    int init_pos = init_offs;
    int init_shift = 1;
    int init_count=0;
    if (istriple(type))
        init_shift = 3;
    
    
    while (cur_node){
        struct tree * cur_init = cur_node -> kids[1];
        if (assigntype( type, cur_init->ty ) != type)
            yyerror("Invalid type in initializer list");
        if (istriple(type) && cur_init->ty == tFLOAT)
            cur_init = node2(tCVTF2TRIPLE, type, cur_init, NULL);
        emit( node2('=', type, constnode(type, init_pos), cur_init));
        init_pos += init_shift;
        cur_node = cur_node->kids[0];
        init_count += 1;
        if (init_count > dim)
            yyerror("Too many initializers in array initializer list");
    }                    
    return 0;
}


int check_list_args(struct tree *p)
{
    int ty = -1;
    while (p != NULL) {
        if (ty == -1)
            ty = p->kids[1]->ty;
        if (p->kids[1]->ty != ty) {
            yyerror("Arguments must have the same type");
        }
        p = p->kids[0];
    }
    return ty;
}



void emit1(unsigned char c)
{
    int new_dot = get_dot();
    code[new_dot] = c;
    set_dot(new_dot+1);
}

void emitshort(int n)
{
    emit1((n>>8)&0xff);
    emit1(n&0xff);
}

void storeshort(int n, int s)
{
    code[n] = (s>>8)&0xff;
    code[n+1] = s&0xff;
}

void emit2(int op, int n)
{
    emit1(op);
    emitshort(n);
}

void emit3(int op, int a, int b)
{
    emit1(op);
    emitshort(a);
    emitshort(b);
}


void notexpression(struct tree *expr)
{
    int child_count = 0;
    int i;
    switch (expr->op){
        case tEQ:
            expr->op = tNE;
            child_count = 2;
            break;
        case tEQ3:
            expr->op = tNE3;
            child_count = 2;
            break;
        case tNE:
            expr->op = tEQ;
            child_count = 2;
            break;
        case tNE3:
            expr->op = tEQ3;
            child_count = 2;
            break;
        case tGT:
            expr->op = tLE;
            child_count = 2;
            break;
        case tLT:
            expr->op = tGE;
            child_count = 2;
            break;
        case tGE:
            expr->op = tLT;
            child_count = 2;
            break;
        case tLE:
            expr->op = tGT;
            child_count = 2;
            break;
        case tOR:
            expr->op = tAND;
            child_count = 2;
            break;
        case tAND:
            expr->op = tOR;
            child_count = 2;
            break;
        case '!':
            child_count = 1;
            break;
        default:
            break;
    }
    for (i=0; i<child_count; i++){
        notexpression(expr->kids[i]);
    }
}

