%{
/****************************************************************************

FILE	gram.y

PURPOSE 
    grammar file for shading language

*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "shade.h"
#include "opcodes.h"
#include "shmem.h"
#include "shdrglob.h"
#include "gener.h"
#include "slgram.h"

int lineno = 1;
int cur_type;
int cur_size;
int cur_offset;
int assign_offset;
int arr_dim_count;
int func_start;
int illum_op;
char *shadername;
int shadertype;
char *cur_func;
int var_type_flags;
int is_iso_shader_ret_val;

#define TYPE_STACK_SIZE  100
#define MAXCODE     (16*1024)
unsigned char code[MAXCODE];
unsigned short type_stack[TYPE_STACK_SIZE];
int type_stack_top;
int func_ret_type;
int func_ret_offset;
int section = S_PARAM;

%}

%union {
    double  f;
    char    *s;
    int i;
    struct tree *e;
    struct patch p;
}

%token tDISPLACEMENT tSURFACE tISO_FUNCTION
%token tFLOAT tCOLOR tPOINT tNORMAL tVECTOR tSTRING
%token tVOID tTRIPLEEXPR tVARYING tUNIFORM tOUTPUT tEXTERN
%token tIF tELSE tWHILE tDO tFOR tBREAK tCONTINUE tRETURN tILLUMINANCE
%token tSCONST tFCONST tID tINDEX tINDEX3
%token tPLUSEQ tMINUSEQ tTIMESEQ tDIVEQ tASSIGNEQ
%token tOR tAND tEQ tEQ3 tNE tNE3 tLT tGT tLE tGE
%token tDU tDV tDERIV tDU3 tDV3 tDERIV3
%token tPI tNULLOP

%token tJMP tSTMT tPRINTF tQU

%token tNOISE tNOISE3 tSIN tASIN tCOS tACOS tTAN tATAN tATAN2 tRADIANS
%token tPNOISE tPNOISE3
%token tDEGREES tSQRT tSQR tINVSQRT tPOW tEXP tLOG tMOD tABS tSIGN tMIN tMIN3
%token tMAX tMAX3 tCLAMP tCEIL tFLOOR tROUND tSTEP tSMOOTHSTEP tSPLINE tSPLINE3
%token tSPLINEARR tSPLINEARR3
%token tMIN3 tMIX3 tMAX3 tCLAMP3 tRANDOM tRANDOM3 tPTLINED tROTATE
%token tCTRANSFORM tCOMP tSETCOMP tCELLNOISE tCELLNOISE3

%token tXCOMP tYCOMP tZCOMP tSETXCOMP tSETYCOMP tSETZCOMP
%token tLENGTH tDISTANCE tAREA tNORMALIZE tFACEFORWARD tREFLECT
%token tREFRACT tFRESNEL tDEPTH tCALCULATENORMAL tMIX
%token tTRANSFORM tVTRANSFORM tNTRANSFORM
%token tAMBIENT tDIFFUSE tPHONG tSPECULAR tSPECULARBRDF tTRACE

%token tATMOSPHERE tLIGHTSOURCE 

%token tCVTF2TRIPLE tADD tADD3 tSUB tSUB3 tMUL tMUL3
%token tDIV tDIV3

%token tCSRGB tCSHSV tCSHSL tCSXYZ tCSXY_Y tCSYIQ tCSEXPR
%token tPSRASTER tPSSCREEN tPSCAMERA tPSWORLD tPSOBJECT tPSSHADER tPSCURRENT tPSNDC tPSEXPR

%token tCATMULL_ROM tBEZIER tBSPLINE tHERMITE tLINEAR tSPLINEEXPR tFUNCTION

%type <f> tFCONST
%type <i> float_const string_const arr_dim shader_type
%type <s> tID tSCONST func_name func_call_name shader_name
%type <i> var_type geomtriple type_def opt_type
%type <e> expr ident builtin assign_expr comma_expr expr_list assignexpr 
%type <e> additive_expr cross_expr multiplicative_expr dot_expr unary_expr primary 
%type <e> opt_expr_list initializer func_call opt_param_list cond_expr 
%type <e> opt_arr_initializer_list func_expr_list arr_expr_list
%type <e> tPI nonc_ident
%type <i> assign_op
%type <p> stmt block if2 locals_stmts patch_list retaddr
%type <p> illum_start
%type <e> rel_expr expr_relation2 expr_relation3 expr_relation4 imm_rel_expr
%type <i> if1 for_init for1 whilehdr wh1 sl1 offset_holder type_holder 
%type <i> arr_offset_holder illum_dot
%type <e> const_colorspace color_space  point_space const_pointspace tFUNCTION

%right '=' tPLUSEQ tMINUSEQ tTIMESEQ tDIVEQ
%left tOR
%left tAND
%left tEQ tNE
%left tLT tGT tLE tGE
%left '+' '-'
%left '*' '/'
%left '^'
%left '.'
%right tUMINUS '!'

%start shader_file

%%


shader_file
        : shader_and_functions
        ;

shader_and_functions
        : /* blank */
        | shader_and_functions shader
        | shader_and_functions function
        ;
shdr1
        : /* */
                { 
                    section = S_PARAM; 
                }
        ;

shdr2   : /* */
                { 
                    section = S_VAR; 
                    emit1(oENDINIT);
                }
        ;

shader_name
        : tID
                {
                    set_context($1);
                    $$ = $1;
                }
        ;

shader  
        : shader_type shader_name shdr1 sl1 '(' parameters ')' shdr2 '{' locals_stmts '}'
                { 
                    if (shadername != NULL){
                        yyerror("Only one shader per file is allowed");
                    }
                    if ($1 == SHADER_TYPE_ISO_FUNCTION && 
                            is_iso_shader_ret_val == 0){
                        yyerror("iso shader requires return value");
                    }
                            
                    shadername = $2;
                    shaderstart = $4;
                    emit1(oEND); 
                }
        ;

shader_type
        : tSURFACE
                {
                    $$ = SHADER_TYPE_SURFACE;
                    shadertype = $$;
                }
        | tISO_FUNCTION
                {
                    $$ = SHADER_TYPE_ISO_FUNCTION;
                    shadertype = $$;
                }
        /*| tDISPLACEMENT*/
        ;

funct1
        : /* blank */
                {
                    section = F_PARAM;
                    emit1(oFSTART);
                }
        ;
funct2
        : /* blank */
                { 
                    section = F_VAR; 
                }
        ;
        
function
        : opt_type func_name sl1 funct1 '(' f_type  parameters ')' 
                    funct2 '{' locals_stmts '}'
                { 
                    struct func_tree * cur_func = get_function($2);
                    backpatch(cur_func->ret_patches, get_dot());
                    cur_func->code_end = get_dot();
                    if ($1 != tVOID && cur_func->func_code == NULL){
                        yyerror("Function requires return value");
                    }
                    emit1(oRET); 
                    remove_context();
                }
        ;

f_type
        : /* blank */
                {  
                    struct _sym *func;
                    func_ret_type=cur_type;
                    func_ret_offset=cur_offset;
                    cur_offset+=cur_size;
                    func = lookup(cur_func);
                    if (func != NULL && (func->flags & F_RET != 0)){
                        yyerror("Duplicate declaration for function");
                    }
                    set_context(cur_func);
                    install_function(cur_func, func_ret_type, func_ret_offset, func_start, 0);
                }
        ;

        
func_name
        : tID
                { 
                    func_start = get_dot();
                    cur_func=$1;
                    $$=$1;
                }    
        ;

                
inline_function
        : type_def  func_name  '(' f_type funct1
                    sl1 parameters ')' funct2  '{' locals_stmts '}'
                {           
                    struct func_tree * cur_func = get_function($2);
                    backpatch(cur_func->ret_patches, get_dot());
                    cur_func->code_end = get_dot();
                    emit1(oRET); 
                    remove_context();
                }
        ;



locals_stmts
        : patch_list
                {
                    backpatch($1.next_patch, get_dot());  
                }
        ;

        
patch_list
        : /* blank */
                { 
                    $$.next_patch=NULL; 
                }
        | patch_list sl1  
                {
                    struct func_tree *func = get_function(get_context_name());
                    if (check_flag)
                        emit1(oCHK);
                    if (func && func->func_code != NULL)
                        yyerror("return should be last statement in function");
                }
        stmt
                {
                    backpatch($1.next_patch, $2); 
                    $$.next_patch= $4.next_patch; 
                }
        | patch_list sl1 param ';'
                {
                    backpatch($1.next_patch, $2);
                    $$.next_patch = NULL; 
                }
        | patch_list sl1 inline_function 
                {
                  backpatch($1.next_patch, get_dot());
                  $$.next_patch = NULL; 
                }
        ;


parameters: /* blank */
        | param_list
        | param_list ';'
        ;

param_list: param
        | param_list ';' param
        ;


param   
        : type_def id_list
                {
                }
        ;
        
type_def
        : 
                {
                    var_type_flags = 0;
                }
         type_modifier var_type
                {
                    $$ = $3;
                }
        ;

id_list
        : id_list2;
        
id_list2
        : id_decl
        | id_list2 ',' id_decl
        ;

id_decl 
        : tID offset_holder initializer
                { 
                    struct _sym *s;                                               
                    struct tree *t;
                    if (var_type_flags & S_EXTERN){
                        s=lookup_prev_decl($1);
                        if( s == NULL){
                            yyerror("Extern variable is not declared\n");
                        }
                        if (s->type != $2){
                            yyerror("different extern type declaration");
                        }
                        add_active_symbol(s);
                        if($3 != NULL){
                            t=node2('=', $2, constnode($2, s->offset), $3);
                            emit(t);
                        }
                    }
                    else{
                        s=lookup($1);
                        if ( s!=NULL && (s->flags & F_PARAM)==0 ){
                            yyerror("duplicate declaration for variable");
                        }
                        s=lookup_cur_context($1);
                        if ( s!=NULL ){
                            warning("declaration hides previous one");
                        }
                        if (section == F_PARAM){
                            struct _sym *new_param = install($1, $2, 
                                assign_offset, section | var_type_flags, 1);
                            
                            struct func_tree *function = get_function(cur_func);
                            add_param(function, new_param);
                        }
                        else {
                            install($1, $2, assign_offset, section|var_type_flags, 1);
                            if ($3 == NULL){
                                if (section == S_PARAM){
                                    yyerror("Shader parameters require default values");
                                }
                            }
                            else{
                                t=node2('=', $2, constnode($2, assign_offset), $3);
                                emit(t);                        
                            }
                        }
                    }
                }
        | tID '[' arr_dim ']' arr_offset_holder opt_arr_initializer_list
                { 
                    struct _sym *s;                                               
                    struct tree *t;
                    int dim = $3;
                    if (dim <=0 && section != F_PARAM)
                        yyerror("Invalid dimension for array declaration.");
                    if (var_type_flags & S_EXTERN){
                        s=lookup_prev_decl($1);
                        if( s == NULL){
                            yyerror("Extern variable is not declared\n");
                        }
                        if (s->type != $5){
                            yyerror("different extern type declaration");
                        }
                        if (s->dim != dim && dim >= 0){
                            yyerror("different dimension in extern declaration");
                        }
                        add_active_symbol(s);
                        gen_arr_initializer($5, dim, assign_offset, $6);
                    }
                    else{
                        s=lookup($1);
                        if ( s!=NULL && (s->flags & F_PARAM)==0 ){
                            yyerror("duplicate declaration for variable");
                        }
                        s=lookup_cur_context($1);
                        if ( s!=NULL ){
                            warning("declaration hides previous one");
                        }
                        if (section == F_PARAM){
                            struct _sym *new_param = install($1, $5, 
                                    assign_offset, section|var_type_flags, dim);
                            struct func_tree *function = get_function(cur_func);
                            add_param(function, new_param);
                        }
                        else {
                            install($1, $5, assign_offset, section|var_type_flags, dim);
                            if (section == S_PARAM){
                                if ($6 == NULL){
                                    yyerror("Shader parameters require default values");
                                }
                            }
                            gen_arr_initializer($5, dim, assign_offset, $6);
                        }
                    }                  
                }                      
        ;

        
arr_offset_holder
        : /* blank */
                {
                    if (arr_dim_count >0){
                        assign_offset=cur_offset;
                        cur_offset += cur_size * arr_dim_count;
                    }
                    else{
                        assign_offset = cur_offset;
                        cur_offset += cur_size;
                    }
                    $$ = cur_type;
                }
        ;

arr_dim
        : /* blank */
                {
                    $$ = -1;
                    arr_dim_count = $$;
                } 
        | tFCONST
                {
                    $$ = (int)$1;
                    arr_dim_count = $$;
                }
        ;


opt_arr_initializer_list
        : /* blank */
                {
                    $$ = NULL;
                }
        | '=' type_holder '{' arr_expr_list '}'
                {
                    $$ = $4;
                }
        ;

arr_expr_list
        : expr
                { $$ = node2(',', $1->ty, NULL, $1); }
        | expr ',' arr_expr_list
                { $$ = node2(',', $1->ty, $3, $1); }
        ;

        
type_holder
        : /* blank */
                { 
                    if (cur_type == tSTRING){
                        $$ = cur_type;
                    }
                    else if (istriple(cur_type) || cur_type == tFLOAT){
                        $$ = cur_type;
                    }
                    else{
                        $$ = tFLOAT;                    
                    }
                }
        ;
                

initializer
        : /* */
                {
                    $$ = NULL; 
                }
        | '=' type_holder  expr
                { 
                    if ($2 == tSTRING){
                        $$ = $3;
                    }
                    else{
                        if (assigntype($2, $3->ty) != $2)
                            yyerror("Illegal type in initialization.");                  
                        if ( istriple( $2) && $3->ty == tFLOAT)        
                            $$ = node2(tCVTF2TRIPLE, $2, $3, NULL);              
                        else
                            $$=$3; 
                    }
                }
        | '=' type_holder string_const
                {
                    $$ = constnode(tSTRING, $3);
                }
        ;

type_modifier
        : /* blank */
        | tVARYING
        | tUNIFORM
        | tOUTPUT type_modifier
                { 
                    var_type_flags |= S_OUTPUT;
                }
        | tEXTERN type_modifier
                { 
                    var_type_flags |= S_EXTERN;
                }
        ;


opt_type
        : /* blank */
                { 
                    $$ = tVOID; 
                }
        | tVOID 
                { 
                    $$ = tVOID; 
                    cur_type = tVOID;
                }
        | var_type
                { 
                    $$ = $1; 
                }
        ;

        
var_type
        : tFLOAT
                {
                    cur_type = tFLOAT;
                    cur_size = 1; 
                    $$ = cur_type; 
                }
        | tCOLOR 
                { 
                    cur_type = tCOLOR;
                    cur_size = 3; 
                    $$ = cur_type; 
                }
        | geomtriple
                {  
                    $$ = $1; 
                    cur_type = $1;
                    cur_size = 3;
                }
        | tSTRING
                {
                    cur_type = tSTRING;
                    cur_size = 0;
                    $$ = cur_type;
                }
        ;


        
geomtriple
        : tPOINT
                {
                    $$ = tPOINT; 
                }
        | tNORMAL
                {
                    $$ = tNORMAL;
                }
        | tVECTOR
                {
                    $$ = tVECTOR;
                }
        ;



        

const_colorspace
        : string_const
                {
                    char *space = getstring($1);
                    if (strcmp(space, "\"rgb\"")==0){
                        $$ = node2(tCSRGB, tCOLOR, NULL, NULL);
                    }
                    else if (strcmp(space, "\"hsv\"")==0){
                        $$ = node2(tCSHSV, tCOLOR, NULL, NULL);
                    }
                    else if (strcmp(space, "\"hsl\"")==0){
                        $$ = node2(tCSHSL, tCOLOR, NULL, NULL);
                    }
                    else if (strcmp(space, "\"xyz\"")==0){
                        $$ = node2(tCSXYZ, tCOLOR, NULL, NULL);
                    }
                    else if (strcmp(space, "\"XYZ\"")==0){
                        $$ = node2(tCSXYZ, tCOLOR, NULL, NULL);
                    }
                    else if (strcmp(space, "\"xyY\"")==0){
                        $$ = node2(tCSXY_Y, tCOLOR, NULL, NULL);
                    }
                    else if (strcmp(space, "\"YIQ\"")==0){
                        $$ = node2(tCSYIQ, tCOLOR, NULL, NULL);
                    }
                    else{
                        yyerror("wrong color space\n");
                    }
                }
        ;

color_space
        : const_colorspace
                {
                    $$ = $1;
                }
        | ident
                {
                    if ($1->ty != tSTRING)
                        yyerror("wrong type for color space");
                    else
                        $$ = node2(tCSEXPR, tCOLOR, $1, NULL);
                }
        ;
        
const_pointspace
        : string_const
                {
                    char *space = getstring($1);
                    if (strcmp(space, "\"raster\"")==0){
                        $$ = node2(tPSRASTER, tPOINT, NULL, NULL);
                    }
                    else if (strcmp(space, "\"screen\"")==0){
                        $$ = node2(tPSSCREEN, tPOINT, NULL, NULL);
                    }
                    else if (strcmp(space, "\"camera\"")==0){
                        $$ = node2(tPSCAMERA, tPOINT, NULL, NULL);
                    }
                    else if (strcmp(space, "\"world\"")==0){
                        $$ = node2(tPSWORLD, tPOINT, NULL, NULL);
                    }
                    else if (strcmp(space, "\"object\"")==0){
                        $$ = node2(tPSOBJECT, tPOINT, NULL, NULL);
                    }
                    else if (strcmp(space, "\"shader\"")==0){
                        $$ = node2(tPSSHADER, tPOINT, NULL, NULL);
                    }
                    else if (strcmp(space, "\"current\"")==0){
                        $$ = node2(tPSCURRENT, tPOINT, NULL, NULL);
                    }
                    else if (strcmp(space, "\"NDC\"")==0){
                        $$ = node2(tPSNDC, tPOINT, NULL, NULL);
                    }
                    else{
                        yyerror("wrong point space\n");
                    }
                }
        ;

point_space
        : const_pointspace
                {
                    $$ = $1;
                }
        | ident
                {
                    if ($1->ty != tSTRING)
                        $$ = NULL;
                    else
                        $$ = node2(tPSEXPR, tPOINT, $1, NULL);
                }
        ;


offset_holder
        : /* blank */
                {
                    if (cur_type == tSTRING){
                        $$ = cur_type;
                        assign_offset = get_string_offset();
                    }
                    else{
                        assign_offset=cur_offset;
                        cur_offset += cur_size;
                        $$ = cur_type;
                    }
                }
        


sl1 
        : /* */
                { $$ = get_dot(); }
        ;

block_incr
        : /* blank */
                { enter_block(); }
        ;
            
block_decr
        : /* blank */
                { leave_block(); }
        ;

block   
        : '{' block_incr locals_stmts block_decr '}'
                { $$.next_patch = $3.next_patch; }
        ;

if1
        : /* */
                { $$ = get_dot(); }
        ;

if2 
        : /* */
                {
                    $$.next_patch = mkilist(get_dot()+1);
                    emit2(oJMP, 0); 
                }
        ;

for_init
        : tFOR '(' expr
                {
                    start_loop();
                    emit($3);
                    $$ = get_dot();
                }
        ;

for1    
        : /* */
                { $$ = get_dot(); }
        ;

whilehdr
        : tWHILE
                {
                    start_loop();
                    $$ = get_dot(); 
                }
        ;

wh1 
        : /* */
                { $$ = get_dot(); }
        ;

illum_start
        : /* blank */
                {
                    $$.true = mkilist(get_dot()+1); emit2(illum_op, 0);
                    $$.false = mkilist(get_dot()+1);    emit2(oJMP, 0);                    
                }
        ;

illum_dot
        : /* blank */
                {
                    $$ = get_dot();
                }
        ;


stmt    
        : tIF '(' imm_rel_expr ')' if1 stmt
                {
                     $$.next_patch = merge($3->tree_patch.false, $6.next_patch); 
                }                
        | tIF '(' imm_rel_expr ')' if1 stmt tELSE if2 if1 stmt
                { 
                    backpatch($3->tree_patch.false, $9);
                    $$.next_patch = merge($6.next_patch, $8.next_patch);
                    $$.next_patch = merge($$.next_patch, $10.next_patch); 
                }
        | for_init ';' imm_rel_expr ';' comma_expr ')' for1 stmt
                { 
                    set_continue();
                    emit($5);
                    $$.next_patch = $3->tree_patch.false;
                    emit2(oJMP, $1); 
                    set_break();
                    end_loop();
                }
        | whilehdr '(' imm_rel_expr ')' wh1 stmt
                {
                    $$.next_patch = $3->tree_patch.false;
                    set_continue();
                    emit2(oJMP, $1);
                    set_break();
                    end_loop();
                }                
        | block
                {
                    $$.next_patch = $1.next_patch;
                }
        | tBREAK ';'
                {
                    struct ilist * stmt = mkilist(get_dot()+1);
                    emit2(oJMP, 0);
                    if (reg_break(stmt, 1)==0){
                        yyerror("Break statement outside of loop");
                    }
                    $$.next_patch = NULL;
                }
        | tBREAK float_const ';'
                {
                    struct ilist * stmt = mkilist(get_dot()+1);
                    int level = (int)getfloat($2);
                    emit2(oJMP, 0);            
                    if (reg_break(stmt, level)==0){
                        yyerror("Break statement count exceeds nested loop count");
                    }
                    $$.next_patch = NULL;
                }
        | tCONTINUE ';'
                {
                    struct ilist * stmt = mkilist(get_dot()+1);
                        emit2(oJMP, 0);            
                    if (reg_continue(stmt, 1)==0){
                        yyerror("Continue statement outside of loop");
                    }
                    $$.next_patch = NULL;
                }
        | tCONTINUE float_const ';'
                {
                    struct ilist * stmt = mkilist(get_dot()+1);
                    int level = (int)getfloat($2);
                    emit2(oJMP, 0);            
                    if (reg_continue(stmt, level)==0){
                        yyerror("Continue statement value exceeds nested loop count");
                    }
                    $$.next_patch = NULL;
                }
        | tRETURN 
                {
                    struct func_tree *cur_func = get_function(get_context_name());
                    if (cur_func != NULL){
                        if (cur_func->type == tVOID){
                            yyerror("return statement in void function");
                        }
                        cur_type = cur_func->type;
                    }
                    else if (shadertype == SHADER_TYPE_ISO_FUNCTION){
                        /* ok, iso shader can return float value */
                    }
                    else{
                        yyerror("return statement not in function scope");
                    }
                }
          expr ';'
                { 
                    struct tree *q = $3;
                    struct func_tree *cur_func = get_function(get_context_name());
                    if (cur_func != NULL){
                        if (assigntype(cur_func->type, q->ty) != cur_func->type)
                            yyerror("Illegal type in return statement");
                        if ( istriple(cur_func->type) && q->ty == tFLOAT)
                            q = node2(tCVTF2TRIPLE, cur_func->type, q, NULL);
                        if (cur_func->func_code == NULL){
                            cur_func->func_code=q;
                        }
                        else{
                            warning("multiple return statements are not supported");
                        }
                    }
                    else if (shadertype == SHADER_TYPE_ISO_FUNCTION){
                        struct tree *rc_expr = NULL;
                        struct tree *iso_val_expr = constnode(tFLOAT, iso_val_DA_OFFS);
                        
                        if (q->ty != tFLOAT){
                            yyerror("iso shader should return float type");
                        }
                        rc_expr = node2('=', tFLOAT, iso_val_expr, q);
                        emit(rc_expr);
                        is_iso_shader_ret_val = 1;
                    }
                    $$.next_patch=NULL;
                }
        | tRETURN 
                {
                    struct func_tree *cur_func = get_function(get_context_name());
                    if (cur_func == NULL){
                        yyerror("return statement not in function scope");
                    }
                    if (cur_func->type != tSTRING){
                        yyerror("wrong return type in function");
                    }
                    cur_type = cur_func->type;
                }
          string_const ';'
                { 
                    struct tree *q = constnode(tSTRING, $3);
                    struct func_tree *cur_func = get_function(get_context_name());
                    if (assigntype(cur_func->type, q->ty) != cur_func->type)
                        yyerror("Illegal type in return statement");
                    if (cur_func->func_code == NULL){
                        cur_func->func_code=q;
                    }
                    else{
                        warning("multiple return statements are not supported");
                    }
                    $$.next_patch=NULL;
                }
        | assignexpr ';'
                {
                     emit($1);
                     $$.next_patch = NULL; 
                }
        | func_call ';'
                { 
                    struct func_tree *function = get_function_by_offset($1->off);
                    if (function == NULL)
                        yyerror("Function not found");
                    check_params(function, $1->kids[1]);
                    insert_function(function->name);
                    $$.next_patch = NULL; 
                    
                }
        | builtin ';'
                { 
                    emit($1);
                    if ($1 -> ty == tFLOAT){
                        emit1(oPOP);
                    }
                    else if (istriple($1 -> ty)){
                        emit1(oPOP3);
                    }
                    $$.next_patch = NULL; 
                }
        | tPRINTF '(' expr opt_expr_list ')' ';'
                { 
                    if ($3->ty != tSTRING){
                        yyerror("printf requires string constant");
                    }
                    if ($4 != NULL)
                        emit($4);
                    emit3(oPRINTF, $3->off, tlen($4));
                    $$.next_patch = NULL; 
                }
        | tPRINTF '(' string_const opt_expr_list ')' ';'
                { 
                    struct tree * str = constnode(tSTRING, $3);
                    if ($4 != NULL)
                        emit($4);
                    emit3(oPRINTF, str->off, tlen($4));
                    $$.next_patch = NULL; 
                }
        | tSETXCOMP '(' nonc_ident ',' expr ')' ';'
                { 
                    if ( isgeomtriple($3->ty) && $5->ty == tFLOAT){
                        struct tree *setcomp = node2(
                            tSETXCOMP, $3->ty, $3, $5);
                        emit(setcomp);
                    }
                    else
                        yyerror("Illegal type for 'setxcomp'");
                    $$.next_patch=NULL;
                }
        | tSETYCOMP '(' nonc_ident ',' expr ')' ';'
                { 
                    if ( isgeomtriple($3->ty) && $5->ty == tFLOAT){
                        struct tree *setcomp = node2(
                            tSETYCOMP, $3->ty, $3, $5);
                        emit(setcomp);
                    }
                    else
                        yyerror("Illegal type for 'setycomp'");
                    $$.next_patch=NULL;
                }
        | tSETZCOMP '(' nonc_ident ',' expr ')' ';'
                { 
                    if ( isgeomtriple($3->ty) && $5->ty == tFLOAT){
                        int offset = install_fconst(2.0);
                        struct tree *setcomp = node2(
                            tSETZCOMP, $3->ty, $3, $5);
                        emit(setcomp);
                    }
                    else
                        yyerror("Illegal type for 'setzcomp'");
                    $$.next_patch=NULL;
                }
        | tSETCOMP '(' nonc_ident ',' expr ',' expr ')' ';'
                { 
                    if ( istriple($3->ty)){
                        if( $5->ty == tFLOAT && $7->ty == tFLOAT){
                            struct tree *setcomp = NULL;
                            if(is_constant($5->off)){
                                double val = getfloat($5->off);
                                if (val == 0.0){
                                    setcomp = node2(tSETXCOMP, $3->ty, $3, $7);
                                }
                                else if (val == 1.0){
                                    setcomp = node2(tSETYCOMP, $3->ty, $3, $7);
                                }
                                else if (val == 2.0){
                                    setcomp = node2(tSETZCOMP, $3->ty, $3, $7);
                                }
                                else{
                                    yyerror("Invalid index value in 'setcomp'");
                                }
                            }
                            else{
                                struct tree *setcomp = node3(
                                    tSETCOMP, $3->ty, $3, $7, $5);
                            }
                            emit(setcomp);
                        }
                        else{
                            yyerror("Illegal index or value for 'setcomp'");
                        }
                    }
                    else
                        yyerror("Illegal variable type for 'setcomp'");
                    $$.next_patch=NULL;
                }
        | tILLUMINANCE '(' expr ')' 
                {
                    emit($3);
                    start_illuminance();
                    illum_op = oILLUM1;
                }
            illum_dot illum_start illum_dot stmt
                {
                    if ( !isgeomtriple($3 -> ty))
                        yyerror("Invalid position expression in 'illuminance'");
                    backpatch($7.true, $8);
                    emit2(oJMP, $6);
                    backpatch($7.false, get_dot());
                    end_illuminance();
                    $$.next_patch = NULL;
                }
        | tILLUMINANCE '(' expr ',' expr ',' expr ')'
                {
                    emit($3);
                    emit($5);
                    emit($7);
                    start_illuminance();
                    illum_op = oILLUM3;
                }
            illum_dot illum_start illum_dot stmt
                {
                    if (!isgeomtriple($3 -> ty) )
                        yyerror("Invalid position in 'illuminance'");
                    if (!isgeomtriple($5 -> ty) )
                        yyerror("Invalid axis expression  in 'illuminance'");
                    if ( $7->ty != tFLOAT)
                        yyerror("Invalid angle expression in 'illuminance'");
                    backpatch($11.true, $12);
                    emit2(oJMP, $10);
                    backpatch($11.false, get_dot());
                    end_illuminance();
                    $$.next_patch = NULL;
                }
        | tFRESNEL '(' expr ',' expr ',' expr ',' nonc_ident ',' nonc_ident ')'
                {
                    struct tree *fresnel_val;
                    if (!isgeomtriple($3->ty))
                        yyerror("Invalid direction component type in 'fresnel'");
                    if (!isgeomtriple($5->ty))
                        yyerror("Invalid  normal component type in 'fresnel'");
                    if ($7->ty != tFLOAT)
                        yyerror("Invalid refraction type in 'fresnel'");
                    if ($9->ty != tFLOAT)
                        yyerror("Invalid reflection coefficent type in 'fresnel'");
                    if ($11->ty != tFLOAT)
                        yyerror("Invalid transmittance coefficent type in 'fresnel'");
                    emit($3);
                    emit($5);
                    emit($7);
                    fresnel_val = node2(tFRESNEL, tFLOAT, $9, $11);
                    emit(fresnel_val);
                    $$.next_patch = NULL;
                }
        | tFRESNEL '(' expr ',' expr ',' expr ',' nonc_ident ',' nonc_ident ',' nonc_ident ',' nonc_ident ')'
                {
                    struct tree *fresnel_val;
                    if (!isgeomtriple($3->ty))
                        yyerror("Invalid direction component type in 'fresnel'");
                    if (!isgeomtriple($5->ty))
                        yyerror("Invalid  normal component type in 'fresnel'");
                    if ($7->ty != tFLOAT)
                        yyerror("Invalid refraction type in 'fresnel'");
                    if ($9->ty != tFLOAT)
                        yyerror("Invalid reflection coefficent type in 'fresnel'");
                    if ($11->ty != tFLOAT)
                        yyerror("Invalid transmittance coefficent type in 'fresnel'");
                    if (!isgeomtriple($13->ty))
                        yyerror("Invalid reflected vector type in 'fresnel'");
                    if (!isgeomtriple($15->ty))
                        yyerror("Invalid transmitted vector type in 'fresnel'");
                    emit($3);
                    emit($5);
                    emit($7);
                    fresnel_val = node4(tFRESNEL, tFLOAT, $9, $11, $13, $15);
                    emit(fresnel_val);
                    $$.next_patch = NULL;
                }
        | ';'
                { $$.next_patch = NULL; }
        ;


retaddr
        : /* blank */
                {
                    $$.next_patch = mkilist(get_dot()+1);
                }
        ;

func_call_name
        : tID
                {
                    $$ = $1;
                }
        ;
        
func_call
        : func_call_name '(' opt_param_list ')' retaddr
                {
                    struct func_tree *function=get_function($1);
                    struct tree *rv;
                    if (function==0){
                        yyerror("function undefined ");
                    }
                    if (function->code_end == 0){
                        yyerror("nested function can not call outer function");
                    }
                    $$ = node2(tFUNCTION, function->type, NULL, $3);
                    $$->off = function->code_start;
                }
        ;

opt_expr_list
        : /* */
                { $$ = NULL; }
        | ',' expr_list
                { $$ = $2; }
        ;
expr_list
        : expr
                { $$ = node2(',', $1->ty, NULL, $1); }
        | expr_list ',' expr
                { $$ = node2(',', $3->ty, $1, $3); }
        ;

opt_param_list
        : /* */
            { $$ = NULL; }
        | func_expr_list
            { $$ = $1; }
        ;

func_expr_list
        : expr
                { $$ = node2('F', $1->ty, NULL, $1); }
        | expr_list ',' expr
                { $$ = node2('F', $3->ty, $1, $3); }
        ;

assign_op: '='
                { $$ = '='; }
        | tTIMESEQ
                { $$ = tTIMESEQ; }
        | tDIVEQ
                { $$ = tDIVEQ; }
        | tPLUSEQ
                { $$ = tPLUSEQ; }
        | tMINUSEQ
                { $$ = tMINUSEQ; }
        ;

comma_expr
        : assign_expr
                { $$ = node2(',', $1->ty, NULL, $1); }
        | comma_expr ',' assign_expr
                { $$ = node2(',', $3->ty, $1, $3); }
        ;

assignexpr
        : nonc_ident assign_op 
                {
                    cur_type = $1->ty;
                }
        assign_expr
                { 
                    struct tree *q;
                    int id_offset;
                    if (assigntype($1->ty, $4->ty) != $1->ty)
                        yyerror("Illegal type in assignment");
                    q = $4;
                    if ( istriple($1->ty) && q->ty == tFLOAT)
                        q = node2(tCVTF2TRIPLE, $1->ty, q, NULL);
                    $$ = node2($2, $1->ty, $1, q);
                }
        ;


primary
        : ident
                { $$ = $1; }
        | float_const
                { $$ = constnode(tFLOAT, $1); }
        | tPI
                { $$ = node2(tPI, tFLOAT, NULL, NULL);}
        | builtin
                {
                    $$ = $1;
                }
        | func_call
                {
                    if ($1 != NULL){
                        $$ = $1;
                    }
                }
        | '('  expr ')'
                { $$ = $2; }
        | '(' expr ',' expr ',' expr ')'
                {
                    if ($2->ty == tFLOAT && $4->ty == tFLOAT && $6->ty == tFLOAT){
                        $$ = node3(tTRIPLEEXPR, cur_type, $6, $4, $2);
                    }
                    else{
                        yyerror("Wrong type for triple expression.\n");
                    }
                }
        ;        
        
unary_expr
        : primary
                { $$ = $1; }
        | tFLOAT
                {
                    type_stack[type_stack_top] = cur_type;
                    if (++type_stack_top == TYPE_STACK_SIZE){
                        yyerror("Type stack overflow.");
                    }
                    cur_type = tFLOAT;
                    cur_size = 1;
                }
          unary_expr
                {
                    struct tree * node=$3;
                    if (node->ty == tFLOAT){
                        if (node->op == tTRIPLEEXPR)
                            yyerror("illegal conversion from triple to float");
                        $$ = node;
                    }
                    else{
                        yyerror("illegal type conversion");
                    }
                    {
                        if (--type_stack_top <0){
                            yyerror("Type stack underflow.");
                        }
                        cur_type = type_stack[type_stack_top];
                    }
                }
        | tCOLOR
                {
                    type_stack[type_stack_top] = cur_type;
                    if (++type_stack_top == TYPE_STACK_SIZE){
                        yyerror("Type stack overflow.");
                    }
                    cur_type = tCOLOR;
                    cur_size = 3;
                }        
          unary_expr
                {
                    struct tree * node=$3;
                    int conv_type = tCOLOR;
                    if (node->ty == tFLOAT){ 
                        if (node->op == tTRIPLEEXPR){
                            warning("type conversion fixup is performed\n");
                            node->ty = conv_type ;
                        }
                        else{
                            node = node2(tCVTF2TRIPLE, tCOLOR, node, NULL);
                        }
                    }
                    if( istriple(node->ty)){
                        node->ty = tCOLOR;
                        $$ = node;
                    }
                    else{
                        yyerror("illegal type conversion");                    
                    }
                    {
                        if (--type_stack_top <0){
                            yyerror("Type stack underflow.");
                        }
                        cur_type = type_stack[type_stack_top];
                    }
                }
        | tCOLOR
                {
                    type_stack[type_stack_top] = cur_type;
                    if (++type_stack_top == TYPE_STACK_SIZE){
                        yyerror("Type stack overflow.");
                    }
                    cur_type = tCOLOR;
                    cur_size = 3;
                }
        
          const_colorspace unary_expr
                {
                    struct tree * node=$4;
                    struct tree * cstype = $3;
                    int conv_type = tCOLOR;
                    if (node->ty == tFLOAT){ 
                        if (node->op == tTRIPLEEXPR){
                            warning("type conversion fixup is performed\n");
                            node->ty = conv_type ;
                        }
                        else{
                            node = node2(tCVTF2TRIPLE, tCOLOR, node, NULL);
                        }
                    }
                    if( istriple(node->ty)){
                        struct tree * tospace = node2(tCSRGB, tCOLOR, NULL, NULL);
                        $$ = node3(tCTRANSFORM, tCOLOR, node, tospace, cstype);
                    }
                    else{
                        yyerror("illegal type conversion");                    
                    }
                    {
                        if (--type_stack_top <0){
                            yyerror("Type stack underflow.");
                        }
                        cur_type = type_stack[type_stack_top];
                    }
                }

        | geomtriple
                {
                    type_stack[type_stack_top] = cur_type;
                    if (++type_stack_top == TYPE_STACK_SIZE){
                        yyerror("Type stack overflow.");
                    }
                    cur_type = $1;
                    cur_size = 3;
                }
          unary_expr
                {
                    struct tree * node=$3;
                    int conv_type = $1;
                    if (node->ty == tFLOAT){ 
                        if (node->op == tTRIPLEEXPR){
                            warning("type conversion fixup is performed\n");
                            node->ty = conv_type ;
                        }
                        else{
                            node = node2(tCVTF2TRIPLE, conv_type, node, NULL);
                        }
                    }
                    if( istriple(node->ty)){
                        node->ty = conv_type;
                        $$ = node;
                    }
                    else{
                        yyerror("illegal type conversion");
                    }
                    {
                        if (--type_stack_top <0){
                            yyerror("Type stack underflow.");
                        }
                        cur_type = type_stack[type_stack_top];
                    }
                }
                
        | geomtriple
                {
                    type_stack[type_stack_top] = cur_type;
                    if (++type_stack_top == TYPE_STACK_SIZE){
                        yyerror("Type stack overflow.");
                    }
                    cur_type = $1;
                    cur_size = 3;
                }        
          const_pointspace unary_expr
                {
                    struct tree * node=$4;
                    struct tree * pstype = $3;
                    int conv_type = $1;
                    if (node->ty == tFLOAT){ 
                        if (node->op == tTRIPLEEXPR){
                            warning("type conversion fixup is performed\n");
                            node->ty = conv_type ;
                        }
                        else{
                            node = node2(tCVTF2TRIPLE, conv_type, node, NULL);
                        }
                    }
                    if( istriple(node->ty)){
                        if (pstype != NULL){
                            struct tree * tospace = node2(tPSCURRENT, conv_type, NULL, NULL);
                            if (conv_type == tPOINT)
                                $$ = node3(tTRANSFORM, conv_type, node, tospace, pstype);
                            else if (conv_type == tNORMAL)
                                $$ = node3(tNTRANSFORM, conv_type, node, tospace, pstype);
                            else
                                $$ = node3(tVTRANSFORM, conv_type, node, tospace, pstype);
                        }
                        else {
                            $$ = node;
                        }
                    }
                    else{
                        yyerror("illegal type conversion");                    
                    }
                    {
                        if (--type_stack_top <0){
                            yyerror("Type stack underflow.");
                        }
                        cur_type = type_stack[type_stack_top];
                    }
                }
        | '-' unary_expr
                {
                    if ($2->ty == tFLOAT) {
                        if (is_constant($2->off)){
                            /* make one expr. from negative number instead of
                                generating number and its negation expr. tree */
                            int offset = $2->off;
                            double val = -getfloat(offset);
                            int new_offset = install_fconst(val);
                            $2->off = new_offset;
                            $$ = $2;
                        }
                        else{
                            $$ = node2(tUMINUS, tFLOAT, $2, NULL);
                        }
                    } 
                    else if ( isgeomtriple($2->ty) ) {
                        $$ = node2(tUMINUS, $2->ty, $2, NULL);
                    } 
                    else {
                        yyerror("Illegal type for unary '-'");
                    }
                }
        ;
 

dot_expr
        : unary_expr
                    { $$ = $1; }
        | dot_expr '.' unary_expr
                {
                    if ( isgeomtriple($1->ty) && isgeomtriple($3->ty) ) {
                        $$ = node2('.', tFLOAT, $1, $3);
                        if ( $1->ty != $3->ty) {
                            warning("different types are used in . expression");
                        }
                    }
                    else if ($1->ty == tCOLOR && $3->ty == tCOLOR) {
                        $$ = node2('.', tFLOAT, $1, $3);
                    }
                    else {
                        yyerror("Illegal types for '.'");
                    }
                }
        ;

multiplicative_expr
        : dot_expr
                { $$ = $1; }
        | multiplicative_expr '*' dot_expr        
                { 
                    if (($$ = arithnode('*', $1, $3)) == NULL)
                        yyerror("Illegal type for binary '*'");
                }
        | multiplicative_expr '/' dot_expr        
                {
                    if (($$ = arithnode('/', $1, $3)) == NULL)
                        yyerror("Illegal type for binary '/'");
                }
        ;


cross_expr
        : multiplicative_expr
                { $$ = $1; }
        | cross_expr '^' multiplicative_expr
                {
                    if ( isgeomtriple($1->ty) && isgeomtriple($3->ty) ) {
                        $$ = node2('^', $1->ty, $1, $3);
                        if ( $1->ty != $3->ty) {
                            warning("different types are used in ^ expression");
                        }
                    }
                    else if ($1->ty == tCOLOR && $3->ty == tCOLOR) {
                        $$ = node2('^', tCOLOR, $1, $3);
                    }
                    else {
                        yyerror("Illegal types for '^'");
                    }
                }
        ;

additive_expr
        : cross_expr
                { $$ = $1; }
        | additive_expr '+' cross_expr
                {
                    if (($$ = arithnode('+', $1, $3)) == NULL)
                        yyerror("Illegal type for binary '+'");
                }
        | additive_expr '-' cross_expr
                {
                    if (($$ = arithnode('-', $1, $3)) == NULL)
                        yyerror("Illegal type for binary '-'");
                }
        ;

expr_relation4
        : '(' rel_expr ')'
                { $$ = $2; }
        | additive_expr tEQ additive_expr
                {
                    $$ = relequiv(tEQ, tEQ3, $1, $3);
                    if ($$ == NULL)
                        yyerror("Illegal type for '=='\n"); 
                }
        | additive_expr tNE additive_expr
                {
                    $$ = relequiv(tNE, tNE3, $1, $3);
                    if ($$ == NULL)
                        yyerror("Illegal type for '!='\n"); 
                }
        | additive_expr tLE additive_expr
                {
                    $$ = relop(tLE, $1, $3);
                    if ($$ == NULL)
                        yyerror("Illegal types for '<='"); 
                }
        | additive_expr tGE additive_expr
                {
                    $$ = relop(tGE, $1, $3);
                    if ($$ == NULL)
                        yyerror("Illegal types for '>='"); 
                }
        | additive_expr tLT additive_expr
                {
                    $$ = relop(tLT, $1, $3);
                    if ($$ == NULL)
                        yyerror("Illegal types for '<'"); 
                }
        | additive_expr tGT additive_expr
                {
                    $$ = relop(tGT, $1, $3);
                    if ($$ == NULL)
                        yyerror("Illegal types for '>'"); 
                }
        ;


expr_relation3
        : '!' expr_relation3 
                {
                    $$ = $2;
                    notexpression($2);
                }
        | expr_relation4
                { $$ = $1; }
        ;
        

expr_relation2
        : expr_relation3 tAND  expr_relation2
                {
                    $$ = node2(tAND, $1->ty, $1, $3);
                }
        | expr_relation3
                { $$ = $1; }
        ;

rel_expr
        : expr_relation2 tOR rel_expr
                {
                    $$ = node2(tOR, $1->ty, $1, $3);
                }
        | expr_relation2
                { $$ = $1; }
        ;


imm_rel_expr
        : rel_expr
                {
                    emit($1);
                    $$ = $1;
                }
        ;
        

        
cond_expr
        : additive_expr
                { $$ = $1; }
        | rel_expr '?' additive_expr ':' cond_expr
                { 
                    $$ = node3(tQU, $3->ty,$1, $3, $5);
                }
        ;

assign_expr
        : assignexpr
                { $$ = $1; }
        | cond_expr
                { $$ = $1; }
        ;


expr    
        : assign_expr
                { $$ = $1; }
        ;

float_const
        : tFCONST
                { $$ = install_fconst($1); }
        ;

string_const
        : tSCONST
                { $$ = install_sconst($1); }
        ;


nonc_ident
        : ident
                {
                    struct _sym * cur_sym;
                    if ($1->ty == tSTRING){
                        cur_sym = get_string_sym($1->off);
                    }
                    else{
                        cur_sym = getsym($1->off);
                    }
                    if ((cur_sym->flags & F_PARAM) && !(cur_sym->flags & S_OUTPUT)){
                        yyerror("Only output parameters could be modified");        
                    }        
                    $$ = $1;
                }
        ;

ident   
        : tID
                { 
                    struct _sym *s;
                    int offset;
                    if ((s = lookup_prev_decl($1)) == NULL)
                        undeclared($1);
                    if ((s->flags & F_ILLUMIN) && !in_illuminance()){
                        yyerror("illuminance variables are available only in illuminance block");
                    }
                    if (get_function(get_context_name()) != NULL && 
                            lookup_cur_context($1) == 0){
                        yyerror("variable not available without extern declaration");
                    }
                    use_sym(s);
                    offset = s->offset;
                    $$ = constnode(s->type, offset); 
                }
        | tID '[' tFCONST ']'
                {
                    struct _sym *s;
                    int offset = $3;
                    if ((s = lookup_prev_decl($1)) == NULL)
                        undeclared($1);
                    if (s->flags & S_BUILTIN){
                        yyerror("Variable is not an array");
                    }
                    if (get_function(get_context_name()) != NULL && 
                            lookup_cur_context($1) == 0){
                        yyerror("variable not available without extern declaration");
                    }
                    if (offset >= s->dim && s->dim >0){
                        /* 
                           negative values represent arrays in function 
                           parameters and are checked during run-time.
                        */
                        yyerror("Array subscription error!");
                    }
                    use_sym(s);
                    if ((s->flags & F_PARAM) != 0){
                        /* 
                            access to formal parameter, dont generate
                            explicit offset access 
                        */
                        int offset = install_fconst($3);
                        struct tree * val = constnode(tFLOAT, offset);
                        if (istriple(s->type))
                            $$ = node2(tINDEX3, s->type, val, NULL);
                        else
                            $$ = node2(tINDEX, s->type, val, NULL); 
                        $$->off = s->offset;
                    }
                    else{
                        if (istriple(s->type))
                            offset *= 3;
                        $$ = constnode(s->type, s->offset + offset);
                    }
                }
        | tID '[' expr ']'
                {
                    struct _sym *s;
                    if ((s = lookup_prev_decl($1)) == NULL)
                        undeclared($1);
                    if (s->flags & S_BUILTIN){
                        yyerror("Variable is not an array");
                    }
                    if (get_function(get_context_name()) != NULL && 
                            lookup_cur_context($1) == 0){
                        yyerror("variable not available without extern declaration");
                    }
                    use_sym(s);
                    if (istriple(s->type))
                        $$ = node2(tINDEX3, s->type, $3, NULL);
                    else
                        $$ = node2(tINDEX, s->type, $3, NULL); 
                    $$->off = s->offset;
                }
        ;

            
builtin 
        : tSIN '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'sin'");
                      $$ = node2(tSIN, tFLOAT, $3, NULL); 
                }
        | tASIN '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'asin'");
                    $$ = node2(tASIN, tFLOAT, $3, NULL); 
                }
        | tCOS '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'cos'");
                    $$ = node2(tCOS, tFLOAT, $3, NULL); 
                }
        | tACOS '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'acos'");
                      $$ = node2(tACOS, tFLOAT, $3, NULL); 
                }
        | tTAN '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'tan'");
                    $$ = node2(tTAN, tFLOAT, $3, NULL); 
                }
        | tATAN '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'atan'");
                    $$ = node2(tATAN, tFLOAT, $3, NULL); 
                }
        | tATAN '(' expr ',' expr ')'
                {
                    if ($3->ty != tFLOAT || $5->ty != tFLOAT)
                        yyerror("Illegal type for 'atan'");
                    $$ = node2(tATAN2, tFLOAT, $3, $5); 
                }
        | tRADIANS '(' expr ')'
                { 
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'radians'");
                    $$ = node2(tRADIANS, tFLOAT, $3, NULL); 
                }
        | tDEGREES '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'degrees'");
                    $$ = node2(tDEGREES, tFLOAT, $3, NULL); 
                }
        | tSQRT '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'sqrt'");
                    $$ = node2(tSQRT, tFLOAT, $3, NULL); 
                }
        | tSQR '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'sqr'");
                    $$ = node2(tSQR, tFLOAT, $3, NULL); 
                }
        | tINVSQRT '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'inversesqrt'");
                    $$ = node2(tINVSQRT, tFLOAT, $3, NULL); 
                }
        | tPOW '(' expr ',' expr ')'
                {
                    if ($3->ty != tFLOAT || $5->ty != tFLOAT)
                        yyerror("Illegal type for 'pow'");
                    $$ = node2(tPOW, tFLOAT, $3, $5); 
                }
        | tEXP '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'exp'");
                    $$ = node2(tEXP, tFLOAT, $3, NULL); 
                }
        | tLOG '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'log'");
                    $$ = node2(tLOG, tFLOAT, $3, NULL); 
                }
        | tLOG '(' expr ',' expr ')'
                {
                    if ($3->ty != tFLOAT || $5->ty != tFLOAT)
                        yyerror("Illegal type for 'log'");
                    $$ = node2(tLOG, tFLOAT, $3, $5); 
                }
        | tMOD '(' expr ',' expr ')'
                {
                    if ($3->ty != tFLOAT || $5->ty != tFLOAT)
                        yyerror("Illegal type for 'mod'");
                    $$ = node2(tMOD, tFLOAT, $3, $5); 
                }
        | tABS '(' expr ')'
                { 
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'abs'");
                    $$ = node2(tABS, tFLOAT, $3, NULL); 
                }
        | tSIGN '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'sign'");
                    $$ = node2(tSIGN, tFLOAT, $3, NULL); 
                }
        | tMIN '(' expr ',' expr opt_expr_list ')'
                { 
                    if ($6 != NULL)
                        check_list_args($6);
                    if ($3->ty == tFLOAT){ 
                        if ($5->ty != tFLOAT)
                            yyerror("Illegal type for 'min'");                  
                        $$ = node3(tMIN, tFLOAT, $3, $5, $6); 
                    }
                    else{
                        $$ = node3(tMIN3, $3->ty, $3, $5, $6); 
                    }
                }                  
        | tMAX '(' expr ',' expr  opt_expr_list')'
                { 
                    if ($6 != NULL)
                        check_list_args($6);
                    if ($3->ty == tFLOAT){ 
                        if ($5->ty != tFLOAT)
                            yyerror("Illegal type for 'max'");                  
                        $$ = node3(tMAX, tFLOAT, $3, $5, $6); 
                    }
                    else{
                        $$ = node3(tMAX3, $3->ty, $3, $5, $6); 
                    }
                }                  
        | tCLAMP '(' expr ',' expr ',' expr')'
                { 
                    if ($3->ty !=  $5->ty || $3->ty != $7->ty )
                        yyerror("Incompatible types for 'clamp'");
                    if ($3->ty == tFLOAT)
                        $$ = node3(tCLAMP, tFLOAT, $3, $5, $7); 
                    else
                        $$ = node3(tCLAMP3, $3->ty, $3, $5, $7); 
                }
        | tCEIL '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'ceil'");
                    $$ = node2(tCEIL, tFLOAT, $3, NULL); 
                }
        | tFLOOR '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                    yyerror("Illegal type for 'floor'");
                    $$ = node2(tFLOOR, tFLOAT, $3, NULL); 
                }
        | tROUND '(' expr ')'
                {
                    if ($3->ty != tFLOAT)
                        yyerror("Illegal type for 'round'");
                    $$ = node2(tROUND, tFLOAT, $3, NULL); 
                }
        | tSTEP '(' expr ',' expr ')'
                {
                    if ($3->ty != tFLOAT || $5->ty != tFLOAT)
                        yyerror("Illegal type for 'step'");
                    $$ = node2(tSTEP, tFLOAT, $3, $5); 
                }
        | tSMOOTHSTEP '(' expr ',' expr ',' expr')'
                {
                    if ($3->ty != tFLOAT || $5->ty != tFLOAT || $7->ty != tFLOAT)
                        yyerror("Illegal type for 'smoothstep'");
                    $$ = node3(tSMOOTHSTEP, tFLOAT, $3, $5, $7); 
                }
        | tXCOMP '(' expr ')'
                {       
                    if ( !isgeomtriple($3->ty) )
                        yyerror("Illegal type for 'xcomp'");
                    $$ = node2(tXCOMP, tFLOAT, $3, NULL); 
                }
        | tYCOMP '(' expr ')'
                {
                    if ( !isgeomtriple($3->ty) )
                        yyerror("Illegal type for 'ycomp'");
                    $$ = node2(tYCOMP, tFLOAT, $3, NULL); 
                }
        | tZCOMP '(' expr ')'
                {
                    if ( !isgeomtriple($3->ty))
                        yyerror("Illegal type for 'zcomp'");
                    $$ = node2(tZCOMP, tFLOAT, $3, NULL); 
                }
        | tCOMP '(' expr ',' expr ')'
                {
                    if ( $3->ty != tCOLOR )
                        yyerror("Illegal type for 1st parameter in  'comp'");
                    if ( $5->ty != tFLOAT )
                        yyerror("Illegal index type for 'comp'");
                    if (is_constant($5->off)){
                        double val = getfloat($5->off);
                        if (val == 0.0){
                            $$ = node2(tXCOMP, tFLOAT, $3, NULL);
                        }
                        else if (val == 1.0){
                            $$ = node2(tYCOMP, tFLOAT, $3, NULL);                        
                        }
                        else if (val == 2.0){
                            $$ = node2(tZCOMP, tFLOAT, $3, NULL);                        
                        }
                        else{
                            yyerror("Illegal index value for 'comp'");
                        }
                    }
                    else{
                        $$ = node2(tCOMP, tFLOAT, $3, $5); 
                    }
                }
        | tLENGTH '(' expr ')'
                {
                    if ( !isgeomtriple($3->ty) )
                        yyerror("Illegal type for 'length'");
                    $$ = node2(tLENGTH, tFLOAT, $3, NULL); 
                }
        | tDISTANCE '(' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty))
                        yyerror("Illegal type for 'distance'");
                    if ( $3->ty != $5->ty)
                        warning("Different types in distance function");
                    $$ = node2(tDISTANCE, tFLOAT, $3, $5); 
                }
        | tNORMALIZE '(' expr ')'
                {
                    if ( !isgeomtriple($3->ty))
                        yyerror("Illegal type for 'normalize'");
                    $$ = node2(tNORMALIZE, tPOINT, $3, NULL); 
                }
        | tFACEFORWARD '(' expr ',' expr ')'
                {
                    struct _sym *Ng = lookup_prev_decl("Ng");
                    struct tree *Ng_expr=constnode(Ng->type, Ng->offset); 
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty))
                        yyerror("Illegal type for 'faceforward'");
                    use_sym(Ng);
                    $$ = node3(tFACEFORWARD, $3->ty, $3, $5, Ng_expr); 
                }
        | tFACEFORWARD '(' expr ',' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty) ||
                         !isgeomtriple($7->ty))
                        yyerror("Illegal type for 'faceforward'");
                    $$ = node3(tFACEFORWARD, $3->ty, $3, $5, $7);
                }
        | tREFLECT '(' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty) )
                        yyerror("Illegal type for 'reflect'");
                    $$ = node2(tREFLECT, $3->ty, $3, $5); 
                }
        | tREFRACT '(' expr ',' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty) || 
                            $7->ty != tFLOAT)
                        yyerror("Illegal type for 'refract'");
                    $$ = node3(tREFRACT, tPOINT, $3, $5, $7); 
                }
        | tMIX '(' expr ',' expr ',' expr ')'
                { 
                    if ( $7->ty != tFLOAT)
                        yyerror("Illegal interpolation value type for 'mix'");
                    if ($3->ty == tFLOAT){
                        if ($5->ty != tFLOAT)
                            yyerror("Incompatible type for second parameter in 'mix'");
                        $$ = node3(tMIX, $3->ty, $3, $5, $7); 
                    }
                    else if (istriple($3->ty)){
                        if ($5->ty == tFLOAT){
                            struct tree * conv = 
                                        node2(tCVTF2TRIPLE, $5->ty, $5, NULL);
                            $$ = node3(tMIX3, $3->ty, $3, $5, $7);
                        }
                        else if (istriple($5->ty)){
                            $$ = node3(tMIX3, $3->ty, $3, $5, $7);
                        }
                        else
                            yyerror("Invalid second parameter type in 'mix'");
                    }
                }
        | tNOISE type_holder '('  expr ')'
                {
                    if ($2 == tFLOAT)
                        $$ = node2(tNOISE, $2, $4, NULL);
                    else if ( istriple($2))
                        $$ = node2(tNOISE3, $2, $4, NULL);
                }
        | tNOISE type_holder '(' expr ',' expr ')'
                {
                    struct tree * tmp = NULL;
                    if ($6 -> ty != tFLOAT)
                        yyerror("Invalid 2nd argument type for 'noise'");

                    if ($2 == tFLOAT){
                        $$ = node2(tNOISE, $2, $4, $6);
                    }
                    else if (istriple($2))
                        $$ = node2(tNOISE3, $2, $4, $6);
                }
        | tCELLNOISE type_holder '(' expr ')'
                {
                    if ( $2 == tFLOAT ){
                        $$ = node2(tCELLNOISE, $2, $4, NULL);
                    }
                    else if ( istriple($2) ){
                        $$ = node2(tCELLNOISE3, $2, $4, NULL);
                    }
                }
        | tCELLNOISE type_holder '(' expr ',' expr ')'
                {                                                               
                    if ($6 -> ty != tFLOAT)                                     
                        yyerror("Invalid 2nd argument type for 'cellnoise'");      
                    if ($2 == tFLOAT){                                      
                        $$ = node2(tCELLNOISE, $2, $4, $6);
                    }                                                           
                    else if (istriple($2))                                 
                        $$ = node2(tCELLNOISE3, $2, $4, $6);                 
                }                                                               
        | tPNOISE type_holder '(' expr ',' expr ')'
                {
                    int result_op;
                    if ($2 == tFLOAT){
                        result_op = tPNOISE;
                    }
                    else if (istriple($2)){
                        result_op = tPNOISE3;
                    }
                    else{
                        yyerror("Invalid conversion in 'pnoise'");
                    }
                    if ($4->ty == tFLOAT){
                        if ($6->ty == tFLOAT){
                            $$ = node2(result_op, $2, $4, $6);
                        }
                        else{
                            yyerror("Incompatible parameter types in 'pnoise'");
                        }
                    }
                    else if (istriple($4->ty)){
                        struct tree *sec_param = $6;
                        if ($6->ty == tFLOAT){
                            sec_param = node2(tCVTF2TRIPLE, $6->ty, $6, NULL);
                            $$ = node2(result_op, $2, $4, sec_param);
                        }
                        else if (istriple($6->ty)){
                            $$ = node2(result_op, $2, $4, $6); 
                        }
                    }
                    else{
                        yyerror("Invalid parameter in 'pnoise'");
                    }
                }
        | tPNOISE type_holder '(' expr ',' expr ',' expr ',' expr ')'
                {
                    int result_op;
                    if ($2 == tFLOAT){
                        result_op = tPNOISE;
                    }
                    else if (istriple($2)){
                        result_op = tPNOISE3;
                    }
                    else{
                        yyerror("Invalid conversion in 'pnoise'");
                    }
                    if ($4->ty == tFLOAT){
                        if ($6->ty == tFLOAT && $8->ty == tFLOAT && $10->ty ==tFLOAT){
                            $$ = node4(result_op, $2, $4, $6, $8, $10);
                        }
                        else{
                            yyerror("Incompatible parameter types in 'pnoise'");
                        }
                    }
                    else if (istriple($4->ty)){
                        struct tree *third = $8;
                        if ($6->ty != tFLOAT || $10->ty != tFLOAT){
                            yyerror("Invalid parameter type in 'pnoise'");
                        }
                        if (third->ty == tFLOAT){
                            third = node2(tCVTF2TRIPLE, third->ty, third, NULL);
                            $$ = node4(result_op, $2, $4, $6, third, $10);
                        }
                        else {
                            $$ = node4(result_op, $2, $4, $6, $8, $10); 
                        }
                    }
                    else{
                        yyerror("Invalid parameter in 'pnoise'");
                    }
                }
        | tTRANSFORM '(' point_space ','  expr ')'
                {
                    struct tree *fromspace, *tospace;
                    if ( !isgeomtriple($5->ty))
                        yyerror("Illegal value in 'transform'");
                    else if ($5->ty != tPOINT)
                        warning("Incompatible type in 'transform'");
                    fromspace = node2(tPSCURRENT, $5->ty, NULL, NULL);
                    tospace = $3;
                    $$ = node3(tTRANSFORM, $5->ty, $5, tospace, fromspace); 
                }
        | tTRANSFORM '(' point_space ',' point_space ','  expr ')'
                {
                    struct tree *fromspace, *tospace;
                    if ( !isgeomtriple($7->ty))
                        yyerror("Illegal value in 'transform'");
                    else if ($7->ty != tPOINT)
                        warning("Incompatible type in 'transform'");
                    fromspace = $3;
                    tospace = $5;
                    $$ = node3(tTRANSFORM, $7->ty, $7, tospace, fromspace); 
                }
        | tNTRANSFORM '(' point_space ','  expr ')'
                {
                    struct tree *fromspace, *tospace;
                    if ( $5->ty != tNORMAL)
                        if (istriple($5->ty))
                            warning("Incompatible type in 'ntransform'");
                        else
                            yyerror("Illegal value in 'ntransform'");
                    fromspace = node2(tPSCURRENT, $5->ty, NULL, NULL);
                    tospace = $3;
                    $$ = node3(tNTRANSFORM, $5->ty, $5, tospace, fromspace); 
                }
        | tNTRANSFORM '(' point_space ',' point_space ','  expr ')'
                {
                    struct tree *fromspace, *tospace;
                    if ( $7->ty != tNORMAL)
                        if (istriple($7->ty))
                            warning("Incompatible type in 'ntransform'");
                        else
                            yyerror("Illegal value in 'ntransform'");
                    fromspace = $3;
                    tospace = $5;
                    $$ = node3(tNTRANSFORM, $7->ty, $7, tospace, fromspace); 
                }
        | tVTRANSFORM '(' point_space ','  expr ')'
                {
                    struct tree *fromspace, *tospace;
                    if ( $5->ty != tVECTOR)
                        if (istriple($5->ty))
                            warning("Incompatible type in 'vtransform'");
                        else
                            yyerror("Illegal value in 'vtransform'");
                    fromspace = node2(tPSCURRENT, $5->ty, NULL, NULL);
                    tospace = $3;
                    $$ = node3(tVTRANSFORM, $5->ty, $5, tospace, fromspace); 
                }
        | tVTRANSFORM '(' point_space ',' point_space ','  expr ')'
                {
                    struct tree *fromspace, *tospace;
                    if ( $7->ty != tVECTOR)
                        if (istriple($7->ty))
                            warning("Incompatible type in 'vtransform'");
                        else
                            yyerror("Illegal value in 'vtransform'");
                    fromspace = $3;
                    tospace = $5;
                    $$ = node3(tVTRANSFORM, $7->ty, $7, tospace, fromspace); 
                }
        | tAMBIENT '(' ')'
                { $$ = node2(tAMBIENT, tCOLOR, NULL, NULL); }
        | tDIFFUSE '(' expr ')'
                {
                    if ( !isgeomtriple($3->ty) )
                        yyerror("Illegal type for 'diffuse'");
                    $$ = node2(tDIFFUSE, tCOLOR, $3, NULL); 
                }
        | tPHONG '(' expr ',' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty) 
                                    || $7->ty != tFLOAT)
                        yyerror("Illegal type for 'phong'");
                    $$ = node3(tPHONG, tCOLOR, $3, $5, $7); 
                }
        | tSPECULAR '(' expr ',' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty) 
                                 || $7->ty != tFLOAT)
                        yyerror("Illegal type for 'specular'");
                    $$ = node3(tSPECULAR, tCOLOR, $3, $5, $7); 
                }
        | tSPECULARBRDF '(' expr ',' expr ',' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty) 
                        || !isgeomtriple($7->ty) ||  $9->ty != tFLOAT)
                        yyerror("Illegal type for 'specularbrdf'");
                    $$ = node4(tSPECULARBRDF, tCOLOR, $3, $5, $7, $9); 
                }
        | tTRACE '(' expr ',' expr ')'
                {
                    if ( !isgeomtriple($3->ty) || !isgeomtriple($5->ty) )
                        yyerror("Illegal type for 'trace'");
                    $$ = node2(tTRACE, tCOLOR, $3, $5); 
                }
        | tSPLINE '(' expr ',' expr_list ')'
                {
                    int ty;
                    int op_type = tSPLINE;
                    struct tree * splinetype;
                    struct tree * cpoints;
                    struct tree * value;
                    if ($3->ty == tSTRING){
                        /* get spline type from first expression,
                           spline value is then list's head  
                           and spline control points are list's tail.
                        */
                        splinetype = node2(tSPLINEEXPR, tFLOAT, $3, NULL);
                        /* find first list element */
                        {
                            struct tree * tmp = $5;
                            while (tmp->kids[0] != NULL && 
                                   tmp->kids[0]-> kids[0] != NULL){
                                tmp = tmp->kids[0];
                            }
                            value = tmp->kids[0]->kids[1];
                            tmp->kids[0] = NULL;
                            cpoints = $5;
                        }
                    }
                    else{
                        splinetype = node2(tCATMULL_ROM, tFLOAT, NULL, NULL);
                        if ($3->ty != tFLOAT)
                            yyerror("Illegal value for 'spline'");
                        value = $3;
                        cpoints = $5;
                    }
                    if (cpoints->kids[0] == NULL && cpoints->kids[1] != NULL &&
                        cpoints->kids[1]->off != -1){
                        /* single identificator, check whether it is array with 
                           enough length
                           -1 indicates function formal parameter, in this case
                            we defer check to run-time
                         */
                        struct _sym * arr_sym = getsym(cpoints->kids[1]->off);
                        if (arr_sym->dim <4 && arr_sym->dim >0)
                            yyerror("spline control point array should contain at least 4 elements!");
                        op_type = tSPLINEARR;
                    }
                    else{
                        int list_len = llen(cpoints);
                        if (list_len < 4)
                            yyerror("spline data should contain at least 4 control points!");
                    }
                    ty = check_list_args(cpoints);
                    if ( istriple(ty )){
                        if (op_type == tSPLINEARR){
                            op_type = tSPLINEARR3;
                        }
                        else {
                            op_type = tSPLINE3;
                        }
                    }
                    $$ = node3(op_type, ty, value, cpoints,  splinetype);
                }

        | tSPLINE '(' string_const ',' expr ',' expr_list ')'
                {
                    int ty;
                    int op_type = tSPLINE;
                    struct tree * splinetype;
                    struct tree * value=$5;
                    struct tree * cpoints=$7;
                    {
                        char * stype = getstring($3);
                        if (strcmp(stype, "\"catmull-rom\"")==0){
                            splinetype = node2(tCATMULL_ROM, tFLOAT, NULL, NULL);
                        }
                        else if (strcmp(stype, "\"bezier\"")==0){
                            splinetype = node2(tBEZIER, tFLOAT, NULL, NULL);
                        }
                        else if (strcmp(stype, "\"bspline\"")==0){
                        splinetype = node2(tBSPLINE, tFLOAT, NULL, NULL);
                        }
                        else if (strcmp(stype, "\"hermite\"")==0){
                            splinetype = node2(tHERMITE, tFLOAT, NULL, NULL);
                        }
                        else if (strcmp(stype, "\"linear\"")==0){
                            splinetype = node2(tLINEAR, tFLOAT, NULL, NULL);
                        }
                        else{
                            yyerror("wrong spline type\n");
                        }
                    }
                    if (cpoints->kids[0] == NULL && cpoints->kids[1] != NULL &&
                        cpoints->kids[1]->off != -1){
                        /* single identificator, check whether it is array with 
                           enough length
                           -1 indicates function formal parameter, in this case
                            we defer check to run-time
                         */
                        struct _sym * arr_sym = getsym(cpoints->kids[1]->off);
                        if (arr_sym->dim <4 && arr_sym->dim >0)
                            yyerror("spline control point array should contain at least 4 elements!");
                        op_type = tSPLINEARR;
                    }
                    else{
                        int list_len = llen(cpoints);
                        if (list_len < 4)
                            yyerror("spline data should contain at least 4 control points!");
                    }
                    ty = check_list_args(cpoints);
                    if ( istriple(ty )){
                        if (op_type == tSPLINEARR){
                            op_type = tSPLINEARR3;
                        }
                        else {
                            op_type = tSPLINE3;
                        }
                    }
                    $$ = node3(op_type, ty, value, cpoints,  splinetype);
                }

        | tCTRANSFORM '(' color_space ',' color_space ',' expr ')'
                {
                    if ($7->ty != tCOLOR){
                        yyerror("color expression expected");
                    }
                    $$ = node3(tCTRANSFORM, tCOLOR, $7, $5, $3);
                }        
        | tCTRANSFORM '(' color_space ',' expr ')'
                {
                    struct tree * fromspace = node2(tCSRGB, tCOLOR, NULL, NULL);
                    if ($5->ty != tCOLOR){
                        yyerror("color expression expected");
                    }
                    $$ = node3(tCTRANSFORM, tCOLOR, $5, $3, fromspace);
                }        
        | tRANDOM type_holder '(' ')'
                { 
                    if ( istriple($2))                                   
                        $$ = node2(tRANDOM3, $2, NULL, NULL);                  
                    else
                        $$ = node2(tRANDOM, $2, NULL, NULL);                   
                }                    
        | tDEPTH '(' expr ')'
                {
                    if (isgeomtriple( $3->ty)){
                        $$ = node2(tDEPTH, tFLOAT, $3, NULL);
                    }
                    else{
                        yyerror(" Illegal argument for 'depth'");
                    }
                }
        | tROTATE '(' expr ',' expr ',' expr ',' expr ')'
                {
                    struct tree * point, *start, *end;
                    if ($5->ty != tFLOAT)
                        yyerror("Wrong angle parameter in 'rotate'");
                    if (!isgeomtriple($3->ty) || !isgeomtriple($7->ty) ||
                        !isgeomtriple($9->ty)){
                        yyerror("Wrong point type in 'rotate'");
                    }
                    $$ = node4(tROTATE, tPOINT, $3, $5, $7, $9);
                }
        | tPTLINED '(' expr ',' expr ',' expr ')'
                {
                    if (isgeomtriple($3->ty) && isgeomtriple($5->ty) && 
                        isgeomtriple($7->ty)){
                        $$ = node3(tPTLINED, tFLOAT, $3, $5, $7);
                    }
                    else {
                        yyerror("Illegal type(s) for 'ptlined'");
                    }
                }
                
                
        | tDU '(' expr ')'
                { 
                    yyerror("Du not implemented");
                    if ($3->ty == tFLOAT){
                        $$ = node2(tDU, $3->ty, $3, NULL);
                    }
                    else if (istriple($3->ty)){
                        $$ = node2(tDU3, $3->ty, $3, NULL);                    
                    }
                }
        | tDV '(' expr ')'
                {
                    yyerror("Dv not implemented");
                    if ($3->ty == tFLOAT){
                        $$ = node2(tDV, $3->ty, $3, NULL);
                    }
                    else if (istriple($3->ty)){
                        $$ = node2(tDV3, $3->ty, $3, NULL);                    
                    }
                }                
        | tDERIV '(' expr ',' expr ')'
                {
                    yyerror("Deriv not implemented");
                    if ($5->ty != tFLOAT){
                        yyerror("Illegal 2nd type in 'Deriv'");
                    }
                    if ($3->ty == tFLOAT){
                        $$ = node2(tDERIV, $3->ty, $3, $5);
                    }
                    else if (istriple($3->ty)){
                        $$ = node2(tDERIV3, $3->ty, $3, $5);                    
                    }
                }   
        | tATMOSPHERE '(' string_const ',' ident ')'
                {
                    yyerror("atmosphere not implemented");
                }
        | tDISPLACEMENT '(' string_const ',' ident ')'
                {
                    yyerror("displacement not implemented");
                }
        | tLIGHTSOURCE '(' string_const ',' ident ')'
                {
                    $$ = node2(tNULLOP, tFLOAT, NULL, NULL);
                    warning("lightsource not implemented");
                }
        | tSURFACE '(' string_const ',' ident ')'
                {
                    yyerror("surface not implemented");
                }
        | tAREA '(' expr ')'
                { 
                    yyerror("area not implemented");
                    if ( $3->ty != tPOINT)
                        yyerror("Invalid parameter for 'area'");
                    $$ = node2(tAREA, tFLOAT, $3, NULL);
                }
        | tCALCULATENORMAL '(' expr ')'
                {
                    yyerror("calculatenormal not implemented");
                    if ( !isgeomtriple($3->ty))
                        yyerror("Illegal type for 'calculatenormal'");
                    $$ = node2(tCALCULATENORMAL, $3->ty, $3, NULL); 
                }
        ;

%%

