/****************************************************************************

FILE	shmem.h

PURPOSE 
    	Memory allocation routines for shading language compiler.
        All allocated blocks are registered here.
        Memory is released in the end of program (by function sh_release_mem),
        as tracking of all allocations during compilation process will require
        considerable effort.

*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/
#include "shmem.h"

#define MEM_BLOCK_COUNT 50

struct mem_block{

    void *ptrs[MEM_BLOCK_COUNT];
    struct mem_block * next_block;
    int ecount;
};

static struct mem_block * root_mem=NULL;

static struct mem_block * alloc_block(){
    int i;
    struct mem_block * mblock = (struct mem_block*) malloc(sizeof (struct mem_block));
    for (i=0; i<MEM_BLOCK_COUNT; i++){
        mblock->ptrs[i] = NULL;
    }
    mblock->next_block = NULL;
    mblock->ecount = 0;
    return mblock;
}

static void register_mem(void *ptr)
{
    int i;
    struct mem_block * mblock=root_mem;
    while (mblock != NULL && mblock->ecount == MEM_BLOCK_COUNT){
        mblock = mblock->next_block;
    }
    if (mblock == NULL){
        mblock = alloc_block();
        mblock->next_block = root_mem;
        root_mem = mblock;
    }
    for (i=0; i<MEM_BLOCK_COUNT; i++){
        if (mblock->ptrs[i]==NULL){
            mblock->ptrs[i]=ptr;
            (mblock->ecount)++;
            break;
        }
    }
}

static void deregister_mem(void * ptr)
{
    struct mem_block * mblock=root_mem;
    int found = 0;
    while (mblock != NULL && !found){
        int i;
        for (i=0; i<MEM_BLOCK_COUNT; i++){
            if (mblock->ptrs[i]==ptr){
                mblock->ptrs[i] = NULL;
                (mblock->ecount)--;
                found = 1;
                break;
            }
        }
        mblock = mblock->next_block;
    }
}


void sh_release_mem(void){

    struct mem_block * mblock=root_mem;
    struct mem_block * prev;
    while (mblock != NULL){
        int i;
        for (i=0; i<MEM_BLOCK_COUNT; i++){
            if (mblock->ptrs[i]!=NULL){
                free(mblock->ptrs[i]);
            }
        }
        prev = mblock;
        mblock = prev->next_block;
        free(prev);
    }
    root_mem = NULL;
}


void * sh_malloc(size_t size)
{

    void * ptr = malloc(size);
    register_mem(ptr);
    return ptr;
}

void sh_free(void * ptr)
{
    deregister_mem(ptr);
    free(ptr);
}

