/****************************************************************************

FILE	shdrglue.h

PURPOSE 
    	This file contains shader declarations, which are used by POV-Ray
*****************************************************************************/
/*****************************************************************************                                                                              
 Copyright (c) 1995
      Robert J. Mercier
 Copyright 2000
      Vahur Krouverk

THIS CODE IS SUPPLIED ``AS IS'' WITH NO WARRANTY EXPRESSED
OR IMPLIED.  NO CLAIMS IS MADE TO IT'S SUITABILITY FOR
ANY PURPOSE AND NO RESPONSIBILITY IS ACCEPTED FOR ANY DAMAGES
OF ANY KIND RESULTING FROM IT'S USE.

When (if) this code is included into POV the usual POV legal stuff will apply.

*****************************************************************************/
#ifndef SHDRGLUE_H
#define SHDRGLUE_H

#include "frame.h"
#include "point.h"  /* for light source declaration */
#include "isosrf.h"

/* This is how arguments are passed from POV to an instance of a shader */
typedef struct _p_arg {
    char    *name;
    short   type;
    short   slot;
    union {
        VECTOR      v;
        DBL         f;
    } v;
    struct _p_arg *next;
} pov_arg_t;

#define T_VECTOR    101
#define T_COLOR     102
#define T_FLOAT     103

#define SHADER_RECURSION_DEPTH 1

/* functions for stats */
int s_nloaded();
int s_ncalls();

enum SHADOW_TYPE {No_Shadow=1, Opaque_Shadow, Os_Shadow, Shader_Shadow};

void Initialize_Shader();

SHADER *Read_Shader_File(char *filename, pov_arg_t *args, enum SHADOW_TYPE TypeOfShadow);
SHADER * Copy_Shader(SHADER * shader);
void Set_Shader_Recursion(int new_depth);
void Use_Shader(SHADER * shader);
void Destroy_Shader(SHADER * shader);

void Exec_Shader ( PIGMENT *Pigment, INTERSECTION *Intersection, RAY *iray,
            VECTOR norm, COLOUR col, VECTOR i_point, int Shadow);
            
DBL Exec_Iso_Shader(FUNCTION *func, VECTOR XYZ);
SHADER *Parse_Iso_Shader(FUNCTION *Function);

#endif

