/****************************************************************************
*                   simcloth.c
*
*
*****************************************************************************/

#include "frame.h"

#ifdef ClothSimPatch

#include "vector.h"
#include "povproto.h"
#include "objects.h"
#include "simcloth.h"



typedef struct {
	int nb1;
	int nb2;
	float nlng;
	float ks;
	VECTOR* points;
	VECTOR* vel;
	VECTOR* constraint;
} CLOTH;


/*****************************************************************************
* Local preprocessor defines
******************************************************************************/
#define VAssign(D,S) {(D)[X] = (S)[X];(D)[Y] = (S)[Y];(D)[Z] = (S)[Z];}
#define VEqEq(A,B) ( ((A)[X] == B[X]) && ((A)[Y] == B[Y]) && ((A)[Z] == B[Z]) )


#define AccessPoint(i,j) (cloth->points[((i)*cloth->nb2)+(j)])
#define AccessVel(i,j)   (cloth->vel[((i)*cloth->nb2)+(j)])
#define AccessConstraint(i,j)   (cloth->constraint[((i)*cloth->nb2)+(j)])
#define AccessResul(i,j) (resul[((i)*cloth->nb2)+(j)])


/*****************************************************************************
* Local typedefs
******************************************************************************/

/*****************************************************************************
* Local typedefs
******************************************************************************/
static const DBL length_mul[3][3] = {
	{0.0, 1.0        , 2.0        },  /* 0,       1,       2 */
	{1.0, 1.414213562, 2.236067977},  /* 1, sqrt(2), sqrt(5) */
	{2.0, 2.236067977, 2.828427125},  /* 2, sqrt(5), sqrt(8) */
};

/*****************************************************************************
* Static functions
******************************************************************************/
static void Iterate_Cloth_Simulation(OBJECT * Object, VECTOR Grav, VECTOR Wind, DBL fric_obj, DBL damp, DBL dt, CLOTH *cloth, VECTOR *resul);

static void Detect_Collision(OBJECT * Object, VECTOR Vel, DBL Disp, VECTOR Start, VECTOR Result, DBL fric_obj);

static int ReadClothFile(FILE *file, CLOTH *cloth);
static int WriteClothFile(FILE *file, CLOTH *cloth);


/*****************************************************************************
* Local variables
******************************************************************************/


/*****************************************************************************
* FUNCTION   Detect_Collision()
******************************************************************************/
static void Detect_Collision(OBJECT * Object, VECTOR Vel, DBL dt, VECTOR Start, VECTOR Result, DBL fric_obj)
{
 INTERSECTION Intersect;
 RAY Ray;
 VECTOR Org;
 VECTOR Norm;
 VECTOR Dir;
 VECTOR IPt;
 VECTOR posDisp;
 DBL dist;
 DBL speed;
 DBL n;

	Initialize_Ray_Containers(&Ray);

	VNormalize(Dir, Vel);
	VAssign(Ray.Direction, Dir);
	VAssign(Ray.Initial, Start);

	if(Intersection(&Intersect, Object, &Ray))
	{
		Assign_Vector(IPt, Intersect.IPoint);
		VLength(speed, Vel);
		VDist(dist, IPt, Start);
		if (dist < (dt*speed))
		{
			Normal(Norm, Intersect.Object, &Intersect);
			VNormalizeEq(Norm);

			/*Set new velocity*/
			VDot(n, Vel, Norm);
			n = -n;
			VAddScaledEq(Vel, n, Norm);
			VScaleEq(Vel, fric_obj);
			VAddScaledEq(Vel, n*0.1, Norm);

			/*Calculate new position*/
			VAddScaled(Result, IPt, (dist*.05), Norm);
			VInverseScale(posDisp, Vel, speed); VScaleEq(posDisp, ((dt*speed)-dist));
			VAddEq(Result, posDisp);

			if (Inside_Object(Result, Object))
			{
				VAddScaled(Result, IPt, (dist*.05), Norm);
			}

			/*Check another collision*/
/*			VLength(dist, posDisp);
			VLength(speed, Vel);
			Detect_Collision(Object, Vel, (dist/speed), IPt, Result, fric_obj);
 */		}
		else
		{
			VAddScaled(Result, Start, dt, Vel);
			if (Inside_Object(Result, Object))
			{
				Assign_Vector(Result, Start);
			}
		}
	}
	else
	{
		VAddScaled(Result, Start, dt, Vel);
		if (Inside_Object(Result, Object))
		{
			Assign_Vector(Result, Start);
		}
	}
}

/*****************************************************************************
* FUNCTION   Iterate_Cloth_Simulation()
******************************************************************************/
static void Iterate_Cloth_Simulation(OBJECT * Object, VECTOR Grav, VECTOR Wind, DBL fric_obj, DBL damp, DBL dt, CLOTH *cloth, VECTOR *resul)
{
 int i, j;
 int k, l;
 VECTOR String;
 VECTOR FricAir;
 VECTOR Force;
 VECTOR Move;
 VECTOR resultmp;
 DBL length;
 DBL ForceScaler;


	for (i=0; i<cloth->nb1; i++)
	{
		/* Calculation of the force exerced by the 24 nearest neighbours */
		for (j=0; j<cloth->nb2; j++)
		{
			/* Gravity */
			VAssign(Force, Grav);
			for(k=-2; k<=2; k++)
			{
				for(l=-2; l<=2; l++)
				{
					if ( ((k != 0) || (l != 0)) &&
					     ((i+k) < cloth->nb1) &&
						 ((i+k) >= 0) &&
						 ((j+l) < cloth->nb2) &&
						 ((j+l) >= 0) )
					{
						VSub(String, AccessPoint(i+k, j+l), AccessPoint(i, j));
						VLength(length, String);
						if (length == 0)
						{
							Make_Vector(String,0,0,0);
						}
						else
						{
							ForceScaler = (cloth->ks)*(length - (length_mul[abs(k)][abs(l)]*(cloth->nlng)))/(length*length);
							VScaleEq(String, ForceScaler);
						}
						VAddEq(Force, String);
					}
				}
			}

			if ((Wind[X]!=0) || (Wind[Y]!=0) || (Wind[Z]!=0))
			{
				/* Calculate Force exerced by the wind */

			}

			/* Update velocity */
			VScaleEq(Force, dt);
			VScaleEq(AccessVel(i, j), damp);
			VAddEq(AccessVel(i, j), Force);
			VScale(Move, AccessVel(i, j), dt);
			/* apply constraint */
			VEvaluateEq(AccessVel(i, j), AccessConstraint(i, j));

			/* Collision dectection */
			if (Object != NULL)
			{
				VLength(length, Move);
				Detect_Collision(Object, AccessVel(i, j), dt, AccessPoint(i, j), resultmp, fric_obj);
			}
			else
			{
				VAdd(resultmp, AccessPoint(i, j), Move);
			}
			Assign_Vector(AccessResul(i,j), resultmp);
		}
	}
}


/*****************************************************************************
* FUNCTION   ReadClothFile()
******************************************************************************/
static int ReadClothFile(FILE *file, CLOTH *cloth)
{
 float x, y, z, vx, vy, vz;
 int i,j;

	if (fscanf(file, "%i, %i, %f, %f,", &(cloth->nb1), &(cloth->nb2),
	                                   &(cloth->nlng), &(cloth->ks)) != 4)
	{
		Status_Info("\nError reading cloth file header...\n");
		return (0);
	}

/*	cloth->nb1 = floor(cloth->nb1);
	cloth->nb2 = floor(cloth->nb2);
*/	if (cloth->nb1 < 2) cloth->nb1 = 2;
	if (cloth->nb2 < 2) cloth->nb2 = 2;
	if (cloth->nlng <= 0) cloth->nlng = 0.0001;
	if (cloth->ks < 0.1) cloth->ks = 0.1;

/*	Status_Info("n1 = %i, n2 = %i, nlng = %f, ks = %f\n", cloth->nb1, cloth->nb2,
	                                   cloth->nlng, cloth->ks);  */

	cloth->points = (VECTOR *)POV_MALLOC(sizeof(VECTOR)*(cloth->nb1)*(cloth->nb2), "cloth points");
	cloth->vel = (VECTOR *)POV_MALLOC(sizeof(VECTOR)*(cloth->nb1)*(cloth->nb2), "cloth points velocity");
	cloth->constraint = (VECTOR *)POV_MALLOC(sizeof(VECTOR)*(cloth->nb1)*(cloth->nb2), "cloth constraints");

	for (i=0; i<cloth->nb1; i++)
	{
		for (j=0; j<cloth->nb2; j++)
		{
			if (fscanf(file, "%f, %f, %f, %f, %f, %f,", &x, &y, &z, &vx, &vy, &vz) == 6)
			{
				AccessPoint(i, j)[X] = x;
				AccessPoint(i, j)[Y] = y;
				AccessPoint(i, j)[Z] = z;
				AccessVel(i, j)[X] = vx;
				AccessVel(i, j)[Y] = vy;
				AccessVel(i, j)[Z] = vz;
			}
			else
			{
				Status_Info("\nError reading cloth points...\n");
				return (0);
			}
		}
	}

	/* constraints init. */
	for (i=0; i<(cloth->nb1*cloth->nb2); i++)
	{
		cloth->constraint[i][X] = 1.0;
		cloth->constraint[i][Y] = 1.0;
		cloth->constraint[i][Z] = 1.0;
	}

/*	Status_Info("\nReading input cloth file constraints...\n"); */
	/* read constraints */
	j=0;
	while (fscanf(file, "%i, %f, %f, %f,", &i, &vx, &vy, &vz) == 4)
	{
	/*	i = floor(i); */
		if (i<((cloth->nb1)*(cloth->nb2)))
		{
			cloth->constraint[i][X] = vx;
			cloth->constraint[i][Y] = vy;
			cloth->constraint[i][Z] = vz;
			j++;
		}
	}

	Status_Info("\n %i constraints read...\n", j);
	return(1);
}

/*****************************************************************************
* FUNCTION   WriteClothFile()
******************************************************************************/
static int WriteClothFile(FILE *file, CLOTH *cloth)
{
 float x, y, z, vx, vy, vz;
 int i,j;

/*	Status_Info("\nWriting output cloth file header...\n"); */
	if (fprintf(file, "%i, %i, %f, %f,\n", cloth->nb1, cloth->nb2,
	                                   cloth->nlng, cloth->ks) < 0)
	{
		Status_Info("Error writing cloth file...\n");
		return (0);
	}

/*	Status_Info("\nWriting output cloth file points...\n"); */
	for (i=0; i<cloth->nb1; i++)
	{
		for (j=0; j<cloth->nb2; j++)
		{
			x = AccessPoint(i, j)[X];
			y = AccessPoint(i, j)[Y];
			z = AccessPoint(i, j)[Z];
			vx = AccessVel(i, j)[X];
			vy = AccessVel(i, j)[Y];
			vz = AccessVel(i, j)[Z];
			if (fprintf(file, "%.12f, %.12f, %.12f, %.12f, %.12f, %.12f,\n", x, y, z, vx, vy, vz) < 0)
			{
				Status_Info("Error writing cloth points...\n");
				return (0);
			}
		}
	}

	/* constraints init. */
	for (i=0; i<((cloth->nb1)*(cloth->nb2)); i++)
	{
		vx = cloth->constraint[i][X];
		vy = cloth->constraint[i][Y];
		vz = cloth->constraint[i][Z];
		if ( (vx != 1.0 ) ||
		     (vy != 1.0 ) ||
		     (vz != 1.0 ) )
		{
			if (fprintf(file, "%i, %.12f, %.12f, %.12f,\n", i, vx, vy, vz) < 0)
			{
				Status_Info("Error writing cloth constraints...\n");
				return (0);
			}
		}
	}

	return(1);
}


/*****************************************************************************
*
* FUNCTION   Make_Cloth_Simulation()
*
*
*
* INPUT
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*
* DESCRIPTION
*
*
* CHANGES
*
*
******************************************************************************/
int Make_Cloth_Simulation(OBJECT * Object, VECTOR Grav, VECTOR Wind, DBL fric_obj, DBL damp, DBL dt, int nbi, char *input, char *output)
{
 FILE *file;
 CLOTH *cloth;
 VECTOR *resul, *temp;
 int i, j;


	/* ================== */
	/* Reading input file */
	/* ================== */

/*	Status_Info("\nOpening cloth file...\n"); */
	if ((file = fopen(input, "r")) == NULL )
	{
		return (0);
	}

	cloth = (CLOTH *)POV_MALLOC(sizeof(CLOTH), "cloth structure");
	cloth->points = NULL;
	cloth->vel = NULL;
	cloth->constraint = NULL;

	if (!ReadClothFile(file, cloth))
	{
		fclose(file);
		POV_FREE(cloth->points);
		POV_FREE(cloth->vel);
		POV_FREE(cloth->constraint);
		return (0);
	}

	/* close file */
	fclose(file);
	resul = (VECTOR *)POV_MALLOC(sizeof(VECTOR)*(cloth->nb1)*(cloth->nb2), "result cloth points");


	/* ====================== */
	/* running the simulation */
	/* ====================== */
	Status_Info("\nRunning cloth simulation...\n");
	for(j=0; j<nbi; j++)
	{
		Status_Info("Simulation %d percent done\r", (int)j*100/nbi);

		COOPERATE_0
		COOPERATE_1

		/* update positions */
		Iterate_Cloth_Simulation(Object, Grav, Wind, fric_obj, damp, dt, cloth, resul);

		/* swap buffers */
		temp = cloth->points;
		cloth->points = resul;
		resul = temp;
	}
	Status_Info("Simulation 100 percent done\n");


	POV_FREE(resul);

	/* =================== */
	/* Writing output file */
	/* =================== */

/*	Status_Info("\nOpening output cloth file...\n"); */
	if ((file = fopen(output, "w")) == NULL )
	{
		POV_FREE(cloth->points);
		POV_FREE(cloth->vel);
		POV_FREE(cloth);
		return (0);
	}

	i = WriteClothFile(file, cloth);
	fclose(file);

	POV_FREE(cloth->points);
	POV_FREE(cloth->vel);
	POV_FREE(cloth);

	if (!i)
	{
		return (0);
	}

	Status_Info("Cloth simulation OK.\n");

	return(1);
}

#endif
